#!/usr/bin/env python3
#
# Copyright (c) Bo Peng and the University of Texas MD Anderson Cancer Center
# Distributed under the terms of the 3-clause BSD License.

import getpass
import os
import time
import base64
from collections import defaultdict

from .utils import env, format_duration, dot_to_gif
from .controller import request_answer_from_controller
from ._version import __version__


class WorkflowSig(object):

    def __init__(self, workflow_id):
        self.data = defaultdict(lambda: defaultdict(list))

        for entry_type, id, item in request_answer_from_controller(
            ['workflow_sig', 'records', workflow_id]):
            try:
                self.data[entry_type][id].append(item.strip())
            except Exception as e:
                env.logger.debug(f'Failed to read report line: {e}')

    def convert_time(self, info):
        for key in [x for x in info.keys() if x.endswith('_time')]:
            info[key + '_str'] = time.strftime('%Y-%m-%d %H:%M:%S',
                                               time.localtime(info[key]))
        if 'start_time' in info and 'end_time' in info:
            info['duration'] = int(info['end_time'] - info['start_time'])
            info['duration_str'] = format_duration(info['duration'])
        return info

    def workflows(self):
        try:
            # workflows has format
            # workflow  workflow_id dict1
            # workflow  workflow_id dict2
            # workflow  workflow_id2 dict1
            workflows = defaultdict(dict)
            for id, values in self.data['workflow'].items():
                for val in values:
                    workflows[id].update(eval(val))
            for v in workflows.values():
                self.convert_time(v)
                if 'dag' in v:
                    try:
                        v['dag_img'] = dot_to_gif(
                            v['dag'], warn=env.logger.warning)
                    except Exception as e:
                        env.logger.warning(
                            f'Failed to obtain convert dag to image: {e}')
                if 'script' in v:
                    v['script'] = base64.b64decode(v['script']).decode()
            return workflows
        except Exception as e:
            env.logger.warning(f'Failed to obtain workflow information: {e}')
            return {}

    def tasks(self):

        def merge_dict(items):
            ret = {}
            for item in items:
                ret.update(eval(item))
            return ret

        try:
            # there can be multiple task status for each id
            tasks = {
                id: merge_dict(res) for id, res in self.data['task'].items()
            }
            for val in tasks.values():
                self.convert_time(val)
                if 'peak_cpu' in val:
                    val['peak_cpu_str'] = f'{val["peak_cpu"]:.1f}%'
                if 'peak_mem' in val:
                    val['peak_mem_str'] = f'{val["peak_mem"] / 1024 / 1024 :.1f}Mb'
            return tasks
        except Exception:
            return {}

    def steps(self):
        try:
            return {
                wf: [self.convert_time(eval(x)) for x in steps
                    ] for wf, steps in self.data['step'].items()
            }
        except Exception as e:
            env.logger.warning(str(e))
            return {}

    def transcripts(self):
        try:
            return {
                step: [self.convert_time(eval(x)) for x in items
                      ] for step, items in self.data['transcript'].items()
            }
        except Exception as e:
            env.logger.warning(str(e))
            return {}

    def placeholders(self):
        try:
            return self.data['placeholder']['file_target']
        except Exception:
            return []


def calc_timeline(info, start_time, total_duration):
    if total_duration == 0 or 'start_time' not in info or 'end_time' not in info:
        info['before_percent'] = 0
        info['during_percent'] = 100
        info['after_precent'] = 0
        return
    info['before_percent'] = int(
        (info['start_time'] - start_time) * 100 / total_duration)
    info['during_percent'] = max(1,
                                 int(info['duration'] * 100 / total_duration))
    info['after_percent'] = 100 - \
        info['before_percent'] - info['during_percent']
    return info


def render_report(output_file, workflow_id):
    data = WorkflowSig(workflow_id)

    from jinja2 import Environment, PackageLoader, select_autoescape
    environment = Environment(
        loader=PackageLoader('sos', 'templates'),
        autoescape=select_autoescape(['html', 'xml']))
    environment.filters['basename'] = os.path.basename
    template = environment.get_template('workflow_report.tpl')

    context = {
        'workflows': data.workflows(),
        'tasks': data.tasks(),
        'steps': data.steps(),
        'transcripts': data.transcripts(),
        'sos_version': __version__,
        'user': getpass.getuser(),
        'time_now_str': time.strftime('%Y-%m-%d %H:%M:%S', time.localtime()),
    }
    # derived context
    context['master_id'] = next(iter(
        context['workflows'].values()))['master_id']
    try:
        # calculate percentage
        start_time = context['workflows'][context['master_id']]['start_time']
        total_duration = context['workflows'][
            context['master_id']]['end_time'] - start_time
        for info in context['workflows'].values():
            calc_timeline(info, start_time, total_duration)
        for steps in context['steps'].values():
            for step in steps:
                calc_timeline(step, start_time, total_duration)
        for info in context['tasks'].values():
            calc_timeline(info, start_time, total_duration)
        with open(output_file, 'w') as wo:
            wo.write(template.render(context))
        env.logger.info(f'Summary of workflow saved to {output_file}')
    except Exception as e:
        env.logger.error(f'Failed to generate report {output_file}: {e}')



#!/usr/bin/env python
#
# Copyright (c) Bo Peng and the University of Texas MD Anderson Cancer Center
# Distributed under the terms of the 3-clause BSD License.

import os
import shutil
import unittest

from sos.parser import SoS_Script
from sos.targets import file_target
from sos.utils import env
from sos.workflow_executor import Base_Executor


class TestActions(unittest.TestCase):

    def setUp(self):
        env.reset()
        self.temp_files = []

    def tearDown(self):
        for f in self.temp_files:
            if file_target(f).exists():
                file_target(f).unlink()

    @unittest.skipIf(not shutil.which('Rscript') or not shutil.which('pandoc'),
                     'R or pandoc not installed')
    def testRmarkdown(self):
        '''Test action Rmarkdown'''
        if file_target('myreport.html').exists():
            file_target('myreport.html').unlink()
        script = SoS_Script(r'''
[10]

report:
## Some random figure

Generated by matplotlib


[100]
# generate report
output: 'myreport.html'
Rmarkdown(output=_output[0])
''')
        wf = script.workflow()
        Base_Executor(wf, config={'report_output': 'report.md'}).run()
        self.assertTrue(os.path.isfile('myreport.html'))
        #
        file_target('myreport.html').unlink()

    @unittest.skipIf(not shutil.which('Rscript') or not shutil.which('pandoc'),
                     'R or pandoc not installed')
    def testRmarkdownWithInput(self):
        # Rmarkdown with specified input.
        script = SoS_Script(r'''
[10]
report: output='a.md'
## Some random figure

Generated by matplotlib


[100]
# generate report
output: 'myreport.html'
Rmarkdown(input='a.md', output=_output[0])
''')
        wf = script.workflow()
        Base_Executor(wf).run()
        self.assertTrue(os.path.isfile('myreport.html'))
        if file_target('myreport.html').exists():
            file_target('myreport.html').unlink()


#     def testRmarkdownWithNoOutput(self):
#         # another case is no output, so output goes to standard output
#         # this cannot be tested in travis because of limit on log file.
#         script = SoS_Script(r'''
# [10]
# report: output='a.md'
# ## Some random figure
#
# Generated by matplotlib
#
#
# [100]
# # generate report
# Rmarkdown(input='a.md')
# ''')
#         wf = script.workflow()
#         Base_Executor(wf).run()

    @unittest.skipIf(not shutil.which('Rscript') or not shutil.which('pandoc'),
                     'R or pandoc not installed')
    def testRmarkdownWithActionOutput(self):
        script = SoS_Script(r'''
[10]
report: output='default_10.md'
A_10

[20]
report: output='default_20.md'
A_20

[100]
# generate report
Rmarkdown(input=['default_10.md', 'default_20.md'], output='output.html')
''')
        wf = script.workflow()
        Base_Executor(wf, config={'report_output': '${step_name}.md'}).run()
        for f in ['default_10.md', 'default_20.md', 'output.html']:
            self.assertTrue(file_target(f).exists())
            file_target(f).unlink()

    @unittest.skipIf(not shutil.which('Rscript') or not shutil.which('pandoc'),
                     'R or pandoc not installed')
    def testRmarkdownToStdout(self):
        script = SoS_Script(r'''
# generate report
Rmarkdown:
    # this is title
''')
        wf = script.workflow()
        Base_Executor(wf).run()

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
#
# Copyright (c) Bo Peng and the University of Texas MD Anderson Cancer Center
# Distributed under the terms of the 3-clause BSD License.

import os

from sos.targets import BaseTarget, textMD5
from sos.utils import env
import shutil


class R_library(BaseTarget):
    '''A target for a R library.'''

    LIB_STATUS_CACHE = {}

    def __init__(self,
                 library,
                 version=None,
                 repos='http://cran.us.r-project.org',
                 autoinstall=False):
        super(R_library, self).__init__()
        self._library = library
        if version is not None:
            version = (version,) if isinstance(version, str) else tuple(version)
        self._version = version
        self._repos = repos
        self._autoinstall = autoinstall

    def _install(self, name, version, repos):
        '''Check existence and version match of R library.
        cran and bioc packages are unique yet might overlap with github.
        Therefore if the input name is {repo}/{pkg} the package will be
        installed from github if not available, else from cran or bioc
        '''
        from sos.pattern import glob_wildcards
        import tempfile
        import subprocess

        output_file = tempfile.NamedTemporaryFile(
            mode='w+t', suffix='.txt', delete=False).name
        script_file = tempfile.NamedTemporaryFile(
            mode='w+t', suffix='.R', delete=False).name
        #
        package_loaded = 'suppressMessages(require(package, character.only=TRUE, quietly=TRUE))'
        version_satisfied = 'TRUE'
        for opt in ('==', '>=', '>', '<=', '<', '!='):
            if opt in name:
                if version is not None:
                    raise ValueError(
                        f"Specifying 'version=' option in addition to '{name}' is not allowed"
                    )
                name, version = [x.strip() for x in name.split(opt, 1)]
                if ',' in version:
                    raise ValueError(
                        f'SoS does not yet support multiple version comparisons. {version} provided'
                    )
                version = (opt + version,)
                break
        if version is not None:
            version = list(version)
            operators = []
            for idx, value in enumerate(version):
                value = str(value)
                if value.startswith('>='):
                    operators.append('>=')
                    version[idx] = value[2:]
                elif value.startswith('>'):
                    operators.append('>')
                    version[idx] = value[1:]
                elif value.startswith('<='):
                    operators.append('<=')
                    version[idx] = value[2:]
                elif value.startswith('<'):
                    operators.append('<')
                    version[idx] = value[1:]
                elif value.startswith('=='):
                    operators.append('==')
                    version[idx] = value[2:]
                elif value.startswith('!='):
                    operators.append('!=')
                    version[idx] = value[2:]
                else:
                    operators.append('==')
            # check version and mark version mismatch
            # if current version satisfies any of the
            # requirement the check program quits
            version_satisfied = '||'.join([
                f'(cur_version {y} {repr(x)})'
                for x, y in zip(version, operators)
            ])
        #
        if len(glob_wildcards('{repo}@{pkg}', [name])['repo']):
            # package is from github
            self._install('remotes', None, repos)
            install_script = f'''
            options(warn=-1)
            package_repo <-strsplit("{name}", split="@")[[1]][2]
            package <-strsplit("{name}", split="@")[[1]][1]
            if ({package_loaded}) cur_version <- packageVersion(package) else cur_version <- NULL
            if (!is.null(cur_version) && {version_satisfied}) {{
                write(paste(package, cur_version, "AVAILABLE"), file={repr(output_file)})
            }} else if ({"TRUE" if self._autoinstall else "FALSE"}) {{
                remotes::install_github(package_repo, force = TRUE)
                if ({package_loaded}) cur_version <- packageVersion(package) else cur_version <- NULL
                # if it still does not exist, write the package name to output
                if (!is.null(cur_version)) {{
                    if ({version_satisfied}) write(paste(package, cur_version, "INSTALLED"), file={repr(output_file)})
                    else write(paste(package, cur_version, "VERSION_MISMATCH"), file={repr(output_file)})
                }} else {{
                    write(paste(package, "NA", "MISSING"), file={repr(output_file)})
                    quit("no")
                }}
            }} else {{
                if (!is.null(cur_version)) write(paste(package, cur_version, "VERSION_MISMATCH"), file={repr(output_file)}) else write(paste(package, cur_version, "UNAVAILABLE"), file={repr(output_file)})
            }}
            '''
        else:
            # package is from cran or bioc
            install_script = f'''
            options(warn=-1)
            package <- "{name}"
            if ({package_loaded}) cur_version <- packageVersion(package) else cur_version <- NULL
            if (!is.null(cur_version) && {version_satisfied}) {{
                write(paste(package, cur_version, "AVAILABLE"), file={repr(output_file)})
            }} else if ({"TRUE" if self._autoinstall else "FALSE"}) {{
                install.packages(package, repos="{repos}", quiet=FALSE)
                # if the package still does not exist
                if (!{package_loaded}) {{
                    source("http://bioconductor.org/biocLite.R")
                    biocLite(package, suppressUpdates=TRUE, suppressAutoUpdate=TRUE, ask=FALSE)
                }}
                if ({package_loaded}) cur_version <- packageVersion(package) else cur_version <- NULL
                # if it still does not exist, write the package name to output
                if (!is.null(cur_version)) {{
                    if ({version_satisfied}) write(paste(package, cur_version, "INSTALLED"), file={repr(output_file)}) else write(paste(package, cur_version, "VERSION_MISMATCH"), file={repr(output_file)})
                }} else {{
                    write(paste(package, "NA", "MISSING"), file={repr(output_file)})
                    quit("no")
                }}
            }} else {{
                if (!is.null(cur_version)) write(paste(package, cur_version, "VERSION_MISMATCH"), file={repr(output_file)}) else write(paste(package, cur_version, "UNAVAILABLE"), file={repr(output_file)})
            }}
            '''
        # temporarily change the run mode to run to execute script
        try:
            with open(script_file, 'w') as sfile:
                sfile.write(install_script)
            #
            p = subprocess.Popen(
                ['Rscript', '--default-packages=utils', script_file])
            ret = p.wait()
            if ret != 0:
                env.logger.warning(
                    f'Failed to detect or install R library {name}')
                return False
        except Exception as e:
            env.logger.error(f'Failed to execute script: {e}')
            return False
        finally:
            os.remove(script_file)

        ret_val = False
        with open(output_file) as tmp:
            for line in tmp:
                lib, cur_version, status = line.split(' ', 2)
                if status.strip() == "MISSING":
                    env.logger.error(
                        f'R library {lib} is not available and cannot be installed.'
                    )
                elif status.strip() == "UNAVAILABLE":
                    env.logger.error(f'R library {lib} is not available.')
                elif status.strip() == 'AVAILABLE':
                    env.logger.debug(
                        f'R library {lib} ({cur_version}) is available')
                    ret_val = True
                elif status.strip() == 'INSTALLED':
                    env.logger.debug(
                        f'R library {lib} ({cur_version}) has been installed')
                    ret_val = True
                elif status.strip() == 'VERSION_MISMATCH':
                    env.logger.error(
                        f'R library {lib} ({cur_version}) does not satisfy version requirement ({"/".join(version)})!'
                    )
                else:
                    raise RuntimeError(f'This should not happen: {line}')
        try:
            os.remove(output_file)
        except Exception:
            pass
        return ret_val

    def target_exists(self, mode='any'):
        if (self._library, self._version,
                self._autoinstall) in self.LIB_STATUS_CACHE:
            return self.LIB_STATUS_CACHE[(self._library, self._version,
                                          self._autoinstall)]
        else:
            # check if R is installed
            if not shutil.which('Rscript'):
                env.logger.debug(
                    f'Target R_Library("{self._library}") does not exist because command Rscript is not found.'
                )
                return False
            ret = self._install(self._library, self._version, self._repos)
            self.LIB_STATUS_CACHE[(self._library, self._version,
                                   self._autoinstall)] = ret
            return ret

    def target_name(self):
        return self._library

    def __repr__(self):
        if self._version:
            return f'{self.__class__.__name__}("{self.target_name()}", {self._version!r})'
        else:
            return super(R_library, self).__repr__()

    def target_signature(self, mode='any'):
        # we are supposed to get signature of the library, but we cannot
        return textMD5(repr(self._library))



#!/usr/bin/env python3
#
# Copyright (c) Bo Peng and the University of Texas MD Anderson Cancer Center
# Distributed under the terms of the 3-clause BSD License.

import os
import subprocess
import sys

import pexpect

from .utils import env
from .targets import path
from .hosts import Host


def list_queues(cfg, hosts=[], verbosity=1):
    env.verbosity = 1
    all_hosts = cfg.get('hosts', [])
    if not all_hosts:
        env.logger.warning(
            "No remote host or task queue is defined in ~/.sos/hosts.yml.")
        return
    for host in hosts:
        if host not in all_hosts:
            env.logger.warning(f'Undefined host {host}')
    host_description = [['Alias', 'Address', 'Queue Type', 'Description'],
                        ['-----', '-------', '----------', '-----------']]
    for host in sorted([x for x in hosts
                        if x in all_hosts] if hosts else all_hosts):
        try:
            h = Host(host, start_engine=False, test_connection=False)
        except Exception as e:
            if verbosity == 0:
                print(f'{host} ({e})')
            elif verbosity in (1, 2):
                host_description.append([host, '?', '?', str(e)])
            else:
                print(f'Queue:       {host}')
                print(f'Error:       {str(e)}')
                if isinstance(cfg['hosts'][host], dict):
                    print('Configuration:')
                    for key in cfg['hosts'][host].keys():
                        print(
                            f'  {(key + ":").ljust(24)} {cfg["hosts"][host][key]}'
                        )
                print()
            continue
        if verbosity == 0:
            print(h.alias)
        elif verbosity in (1, 2):
            host_description.append(
                [h.alias, h._host_agent.address, h._engine_type, h.description])
        else:
            print(f'Queue:       {h.alias}')
            print(f'Address:     {h._host_agent.address}')
            print(f'Queue Type:  {h._engine_type}')
            print(f'Description: {h.description}')
            print('Configuration:')
            keys = sorted(h.config.keys())
            for key in keys:
                print(f'  {(key + ":").ljust(24)} {h.config[key]}')
            print()
    if verbosity in (1, 2):
        width = [(len(x) for x in row) for row in host_description]
        max_width = [max(x) for x in zip(*width)]
        print('\n'.join(' '.join([t.ljust(w)
                                  for t, w in zip(row, max_width)])
                        for row in host_description))


def status_of_queues(cfg, hosts=[], verbosity=1):
    env.verbosity = 1
    all_hosts = cfg.get('hosts', [])
    if not all_hosts:
        env.logger.warning(
            "No remote host or task queue is defined in ~/.sos/hosts.yml.")
        return
    for host in hosts:
        if host not in all_hosts:
            env.logger.warning(f'Undefined host {host}')
    host_description = [[
        'Alias', 'Address', 'Queue Type', 'Running', 'Pending', 'Completed'
    ], ['-----', '-------', '----------', '-------', '-------', '---------']]
    for host in sorted([x for x in hosts
                        if x in all_hosts] if hosts else all_hosts):
        try:
            h = Host(host, start_engine=True)
            status = h._task_engine.query_tasks(tasks=[], verbosity=0)
            if not status:
                raise ValueError("Failed to query status")
        except Exception as e:
            if verbosity == 0:
                print(f'{host} ({e})')
            elif verbosity in (1, 2):
                host_description.append([host, '?', '?', '?', '?', '?'])
            else:
                print(f'Queue:       {host}')
                print(f'Error:       {str(e)}')
                if isinstance(cfg['hosts'][host], dict):
                    print('Configuration:')
                    for key in cfg['hosts'][host].keys():
                        print(
                            f'  {(key + ":").ljust(24)} {cfg["hosts"][host][key]}'
                        )
                print()
            continue
        status = [x.strip() for x in status.splitlines() if x.strip()]
        running = str(status.count('running'))
        pending = str(status.count('pending'))
        completed = str(status.count('completed'))

        if verbosity == 0:
            print(f'{h.alias} {running} {pending} {completed}')
        elif verbosity in (1, 2):
            host_description.append([
                h.alias, h._host_agent.address, h._engine_type, running,
                pending, completed
            ])
        else:
            print(f'Queue:       {h.alias}')
            print(f'Address:     {h._host_agent.address}')
            print(f'Queue Type:  {h._engine_type}')
            print(f'Description: {h.description}')
            for k in set(status):
                if k not in ('running', 'pending', 'completed'):
                    print(f'{(k+":").ljust(26)} {status.count(k)}')
            print('Configuration:')
            keys = sorted(h.config.keys())
            for key in keys:
                print(f'  {(key + ":").ljust(24)} {h.config[key]}')
            print()
    if verbosity in (1, 2):
        width = [(len(x) for x in row) for row in host_description]
        max_width = [max(x) for x in zip(*width)]
        print('\n'.join(' '.join([t.ljust(w)
                                  for t, w in zip(row, max_width)])
                        for row in host_description))


def test_ssh(host):
    return host.test_connection()


def test_scp(host):
    if host.address == 'localhost':
        return 'OK'
    # test send task file
    import random
    tID = random.randint(1, 100000)
    task_filename = os.path.join(
        os.path.expanduser('~'), '.sos', f'test_{tID}.tmp')
    with open(task_filename, 'w') as test_task:
        test_task.write('test task')
    # test scp
    try:
        host.send_job_file(task_filename)
    except Exception as e:
        return str(e)
    # test remove file using ssh
    try:
        host.check_call(f'rm -f ~/.sos/test_{tID}.tmp')
    except Exception as e:
        return str(e)
    # test rsync
    return 'OK'


def test_cmd(host, cmd):
    # test the execution of sos commands
    try:
        ret = host.check_call(
            cmd, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
        if ret == 0:
            return 'OK'
        else:
            return 'sos not installed or not accessible on host.'
    except Exception as e:
        return str(e)


def test_paths(host):
    if host.address == 'localhost':
        return 'OK'
    # shared means, if localhost creates a file, it should be
    # instantly available on the remote host
    if not host.path_map:
        return 'No path_map between local and remote host.'
    import random
    tID = random.randint(1, 100000)
    for local, remote in host.path_map.items():
        if local in host.shared_dirs:
            # will be tested by 'shared'
            continue
        # now, let us see if two directory has the same files?
        if not os.path.isdir(local):
            return f'Mapped directory {local} does not exist.'
        # remote?
        try:
            host.check_output(f'ls -a {path(remote):q}')
        except Exception:
            return f'Failed to access shared directory {remote} on remote host.'

        # test if local directory is writable
        try:
            with open(os.path.join(local, f".sos_test_{tID}.txt"),
                      'w') as tFile:
                tFile.write(f'{tID}')
        except Exception:
            return f'Failed to write to mapped directory {local}'
        # test if file can be sent
        try:
            host.send_to_host(os.path.join(local, f".sos_test_{tID}.txt"))
        except Exception as e:
            return f'Failed to send files under {local} to remote host under {remote}: {e}'
        # the file should be available on remote host
        try:
            remote_content = host.check_output(
                f'cat {remote}/.sos_test_{tID}.txt')
        except Exception as e:
            return f'Failed to send files under {local} to remote host under {remote}: {e}'
        if remote_content != str(tID):
            return f'Content of file sent does not match: {tID} sent, {remote_content} received'
        # test retrieving files
        # remove local file
        os.remove(os.path.join(local, f".sos_test_{tID}.txt"))
        # copy file back
        try:
            host.receive_from_host(os.path.join(local, f".sos_test_{tID}.txt"))
        except Exception as e:
            return f'Failed to receive file from remote host {remote}: {e}'
        #
        if not os.path.isfile(os.path.join(local, f".sos_test_{tID}.txt")):
            return f'Failed to receive file from remote host {remote}: file does not exist'
        # check file content?
        with open(os.path.join(local, f".sos_test_{tID}.txt"), 'r') as tFile:
            remote_content = tFile.read()
        if remote_content != str(tID):
            return f'Content of received file does not match: {tID} expected, {remote_content} received.'
        # if everything ok, remove local and remote test files
        os.remove(os.path.join(local, f".sos_test_{tID}.txt"))
        #
        try:
            remote_content = host.check_output(
                f'rm {remote}/.sos_test_{tID}.txt')
        except Exception as e:
            return f'Failed to remove test file on remote host: {e}'
    return 'OK'


def test_shared(host):
    if host.address == 'localhost':
        return 'OK (localhost)'
    # shared means, if localhost creates a file, it should be
    # instantly available on the remote host
    for local in host.shared_dirs:
        if local not in host.path_map:
            return f'shared directory {local} not in path_map'
        # now, let us see if two directory has the same files?
        if not os.path.isdir(local):
            return f'shared directory {local} does not exist.'
        local_files = os.listdir(local)
        # remote?
        remote = host.path_map[local]
        try:
            remote_files = host.check_output(f'ls -a {path(remote):q}')
        except Exception:
            return f'Failed to access shared directory {remote} on remote host.'
        remote_files = [
            x for x in remote_files.splitlines() if x not in ('.', '..')
        ]
        #
        if sorted(local_files) != sorted(remote_files):
            return f'shared directory {local} has different content on remote host under {remote}'

    return f'OK (shared {" ".join(host.shared_dirs)})'


def stty_sane():
    try:
        subprocess.check_call('stty sane', shell=True)
    except Exception:
        pass


def test_queue(host, cmd=None):
    try:
        h = Host(host, start_engine=False)
    except Exception:
        return [host, '?', '?', '-', '-', '-', '-', '-']
    ssh_res = test_ssh(h._host_agent)
    return [
        h.alias, h._host_agent.address, h._engine_type, ssh_res,
        test_scp(h._host_agent) if ssh_res.startswith('OK') else '-',
        test_cmd(h._host_agent, ["sos", "-h"])
        if ssh_res.startswith('OK') else '-',
        test_paths(h._host_agent) if ssh_res.startswith('OK') else '-',
        test_shared(h._host_agent) if ssh_res.startswith('OK') else '-'
    ] + ([] if cmd is None else [test_cmd(h._host_agent, cmd)])


def test_queues(cfg, hosts=[], cmd=None, verbosity=1):
    env.verbosity = 1
    all_hosts = cfg.get('hosts', [])
    if not all_hosts:
        env.logger.warning(
            "No remote host or task queue is defined in ~/.sos/hosts.yml.")
        return
    host_description = [[
        'Alias', 'Address', 'Queue Type', 'ssh', 'scp', 'sos', 'paths', 'shared'
    ] + ([] if cmd is None else [" ".join(cmd)]),
                        [
                            '-----', '-------', '----------', '---', '---',
                            '---', '-----', '------'
                        ] + ([] if cmd is None else ['-' * len(" ".join(cmd))])]
    for host in hosts:
        if host not in all_hosts:
            env.logger.warning(f'Undefined host {host}')
    hosts = [x for x in hosts if x in all_hosts] if hosts else all_hosts
    if not hosts:
        return
    from multiprocessing import Pool
    pool = Pool(min(len(hosts), 10))
    host_description.extend(pool.starmap(test_queue, [(x, cmd) for x in hosts]))
    if verbosity == 0:
        # just print succ or self
        for hd in host_description:
            print(f'{hd[0]} {"OK" if all(x=="OK" for x in hd[3:]) else "FAIL"}')
    elif verbosity in (1, 2):
        shortened = host_description[:2]
        for row in host_description[2:]:
            shortened.append(row[:3] + [
                'OK' if x.startswith('OK') else ('-' if x == '-' else 'FAIL')
                for x in row[3:]
            ])
        width = [(len(x) for x in row) for row in shortened]
        max_width = [max(x) for x in zip(*width)]
        print('\n'.join(' '.join([t.ljust(w)
                                  for t, w in zip(row, max_width)])
                        for row in shortened))
        if any('FAILED' in row for row in shortened):
            print(
                '\nUse command "sos remote --test host -v3" to check details of hosts with failed tests.'
            )
    else:
        for row in host_description[2:]:
            print(f'Alias:       {row[0]}')
            print(f'Address:     {row[1]}')
            print(f'Queue Type:  {row[2]}')
            print(f'ssh:         {row[3]}')
            print(f'scp:         {row[4]}')
            print(f'sos:         {row[5]}')
            print(f'paths:       {row[6]}')
            print(f'shared:      {row[7]}')
            if cmd:
                print(
                    f'{" ".join(cmd)}:{" "*(max(1, 12 - len(" ".join(cmd))))}{row[8]}'
                )
            print()


def copy_public_key(host, agent, password):
    try:
        if password is None:
            import getpass
            password = getpass.getpass(
                f'Please enter password for {agent.address}: ')
        cmd = f"scp -P {agent.port if agent.port else 22} {os.path.expanduser('~')}/.ssh/id_rsa.pub {agent.address}:id_rsa.pub.{host}"
        env.logger.info(cmd)
        p = pexpect.spawn(cmd, echo=False)
        i = p.expect([
            "(?i)are you sure you want to continue connecting", "[pP]assword:",
            pexpect.EOF
        ])
        if i == 0:
            p.sendline('yes')
            p.expect([
                "(?i)are you sure you want to continue connecting",
                "[pP]assword:", pexpect.EOF
            ],
                     timeout=5)

        if i == 1:
            p.sendline(password)
            i = p.expect(['assword:', pexpect.EOF])
            if i == 0:
                p.close(force=True)
                return f'Incorrect password specified (you can try to specify it with command line option --password)'
        if i == 2:
            p.close()
            return f'Failed to copy public key to {agent.address}'
    except Exception as e:
        p.close()
        return f'Failed to copy public key to {host}: {e}'
    #
    # ssh
    try:
        cmd = f"ssh {agent.address} -p {agent.port} '[ -d .ssh ] || mkdir .ssh && chmod 700 .ssh; cat id_rsa.pub.{host} >> .ssh/authorized_keys; rm -f id_rsa.pub.{host}'"
        env.logger.info(cmd)
        p = pexpect.spawn(cmd, echo=False)
        i = p.expect([
            "(?i)are you sure you want to continue connecting", 'assword:',
            pexpect.EOF
        ])
        if i == 0:
            p.sendline('yes')
            p.expect([
                "(?i)are you sure you want to continue connecting",
                "[pP]assword:", pexpect.EOF
            ],
                     timeout=5)

        if i == 1:
            p.sendline(password)
            i = p.expect(['assword:', pexpect.EOF])
            if i == 0:
                p.close(force=True)
                return f'Incorrect password specified (you can try to specify it with command line option --password)'
        elif i != 1:
            p.close()
            return f'Failed to append public key to .ssh/authorized_keys'
    except Exception as e:
        p.close()
        return f'Failed to append public key to .ssh/authorized_keys: {e}'
    return 'OK'


def setup_remote_access(cfg, hosts=[], password='', verbosity=1):
    env.verbosity = verbosity
    all_hosts = cfg.get('hosts', [])
    if not all_hosts and not hosts:
        env.logger.warning(
            "No remote host or task queue is defined in ~/.sos/hosts.yml.")
        return
    for host in hosts:
        if host not in all_hosts:
            env.logger.warning(
                f'Treating undefined host {host} as address of a remote host.')
    # public_key
    public_key = os.path.join(os.path.expanduser('~'), '.ssh', 'id_rsa.pub')

    from argparse import Namespace
    for host in sorted(hosts if hosts else all_hosts):
        try:
            if host in all_hosts:
                h = Host(host, start_engine=False, test_connection=False)
                host_agent = h._host_agent
                #
                # can ssh already?
                response = test_ssh(host_agent)
                if response.startswith('OK'):
                    env.logger.info(
                        f'Public key access is already enabled for host ``{host}`` with address ``{host_agent.address}``'
                    )
                    continue
                elif 'Could not resolve hostname' in response:
                    env.logger.error(response)
                    sys.exit(1)
            else:
                host_agent = Namespace(address=host, port=22)
        except Exception as e:
            env.logger.error(f'Failed to start task engine {host}: {e}')
            continue

        if os.path.isfile(public_key):
            env.logger.info('Using existing public key .ssh/id_rsa.pub')
        else:
            env.logger.info(f'Public key {public_key} is found. Creating one.')
            try:
                subprocess.check_output(
                    'echo | ssh-keygen -t rsa',
                    shell=True,
                    stderr=subprocess.STDOUT).decode()
                if not os.path.isfile(public_key):
                    raise RuntimeError('public key not found after ssh-keygen')
            except subprocess.CalledProcessError as e:
                print(e.output.decode())
                raise RuntimeError(f'Failed to create public key: {e}')
        #
        response = copy_public_key(host, host_agent, password)
        stty_sane()
        if not response.startswith('OK'):
            env.logger.error(response)
            sys.exit(1)
        # file copied, check ssh again.
        if isinstance(host_agent, Namespace):
            host_agent = Host(host, start_engine=False, test_connection=False)._host_agent
        response = test_ssh(host_agent)
        if response.startswith('OK'):
            env.logger.info(
                f'Public key access is successfully set up for host ``{host}`` with address ``{host_agent.address}``'
            )
            continue
        else:
            env.logger.error(
                f'Failed to connect to {host} after passing public key. Possible problems include permission of .ssh and home directories.'
            )


def login_host(cfg, host):
    try:
        h = Host(host, start_engine=False)
    except Exception as e:
        raise ValueError(f'Unrecognized or invalid host {host}: {e}')

    address, port = h._host_agent.address, h._host_agent.port
    try:
        env.logger.info(f'Running ``ssh {address} -p {port}``')
        os.execvp('ssh', ['ssh', address, '-p', str(port)])
    except Exception as e:
        raise RuntimeError(f'Failed to log in to {host}: {e}')


def run_command_on_hosts(cfg, hosts, cmd, verbosity):
    env.verbosity = verbosity
    if not hosts:
        hosts = cfg.get('hosts', [])
    if not hosts:
        env.logger.warning(
            "No remote host or task queue is defined in ~/.sos/hosts.yml.")
        return
    for host in hosts:
        # runing command on all hosts
        try:
            env.logger.info(f'Running ``{" ".join(cmd)}`` on ``{host}``')
            h = Host(host, start_engine=False)
            print(h._host_agent.check_output(cmd, under_workdir=True))
        except Exception as e:
            from .utils import get_traceback
            if verbosity and verbosity > 2:
                sys.stderr.write(get_traceback())
            env.logger.error(str(e))


def push_to_hosts(cfg, hosts, items, verbosity):
    env.verbosity = verbosity
    if not hosts:
        hosts = cfg.get('hosts', [])
    if not hosts:
        env.logger.warning(
            "No remote host or task queue is defined in ~/.sos/hosts.yml.")
        return
    for host in hosts:
        try:
            env.logger.info(f'Pushing ``{" ".join(items)}`` to ``{host}``')

            h = Host(host, start_engine=False)
            #
            sent = h.send_to_host(items)
            #
            print('{} item{} sent:\n{}'.format(
                len(sent), ' is' if len(sent) <= 1 else 's are', '\n'.join([
                    '{} => {}'.format(x, sent[x]) for x in sorted(sent.keys())
                ])))
        except Exception as e:
            from .utils import get_traceback
            if verbosity and verbosity > 2:
                sys.stderr.write(get_traceback())
            env.logger.error(str(e))
            sys.exit(1)


def pull_from_host(cfg, hosts, items, verbosity):
    env.verbosity = verbosity
    if not hosts:
        hosts = cfg.get('hosts', [])
    if not hosts:
        env.logger.warning(
            "No remote host or task queue is defined in ~/.sos/hosts.yml.")
        return
    if len(hosts) > 1:
        raise ValueError('Can only pull from a single remote host.')
    try:
        env.logger.info(f'Pulling ``{" ".join(items)}`` from ``{hosts[0]}``')

        host = Host(hosts[0], start_engine=False)
        #
        received = host.receive_from_host(items)
        #
        print('{} item{} received:\n{}'.format(
            len(received), ' is' if len(received) <= 1 else 's are', '\n'.join([
                '{} <= {}'.format(x, received[x])
                for x in sorted(received.keys())
            ])))
    except Exception as e:
        from .utils import get_traceback
        if verbosity and verbosity > 2:
            sys.stderr.write(get_traceback())
        env.logger.error(str(e))
        sys.exit(1)



#!/usr/bin/env python3
#
# Copyright (c) Bo Peng and the University of Texas MD Anderson Cancer Center
# Distributed under the terms of the 3-clause BSD License.

import os
import socket
import sys
import unittest

from sos.eval import accessed_vars, on_demand_options
from sos.parser import SoS_Script
from sos.pattern import expand_pattern, extract_pattern
from sos.targets import executable, sos_targets, file_target, sos_step
# these functions are normally not available but can be imported
# using their names for testing purposes
from sos.utils import WorkflowDict, env, get_logger, stable_repr, split_fstring, as_fstring
from sos.workflow_executor import analyze_section
from sos.workflow_executor import Base_Executor


def internet_on(host='8.8.8.8', port=80, timeout=3):
    '''Test if internet is connected '''
    try:
        socket.setdefaulttimeout(timeout)
        socket.socket(socket.AF_INET, socket.SOCK_STREAM).connect((host, port))
        return True
    except Exception as e:
        print(e)
        return False


with_network = internet_on()


class TestUtils(unittest.TestCase):

    def setUp(self):
        env.reset()

    def testLogger(self):
        '''Test logging level'''
        for verbosity in [0, 1, 2, 3, 4]:
            env.verbosity = verbosity
            get_logger().debug(
                'Verbosity {}:debug message with ``empahsized text`` in between'
                .format(env.verbosity))
            get_logger().info(
                'Verbosity {}:info message with ``empahsized text`` in between'
                .format(env.verbosity))
            get_logger().warning(
                'Verbosity {}:warning message with ``empahsized text`` in between'
                .format(env.verbosity))
            get_logger().error(
                'Verbosity {}:error message with ``empahsized text`` in between'
                .format(env.verbosity))

    def testWorkflowDict(self):
        '''Test workflow dict with attribute access'''
        env.reset()
        d = WorkflowDict()
        d['a'] = 1
        self.assertEqual(d['a'], 1)
        #
        d['a'] += 1
        self.assertEqual(d['a'], 2)

    def testPatternMatch(self):
        '''Test snake match's pattern match facility'''
        res = extract_pattern('{a}-{b}.txt', ['file-1.txt', 'file-ab.txt'])
        self.assertEqual(res['a'], ['file', 'file'])
        self.assertEqual(res['b'], ['1', 'ab'])
        res = extract_pattern('{a}-{b}.txt', ['file--ab--cd.txt'])
        self.assertEqual(res['a'], ['file--ab-'])
        self.assertEqual(res['b'], ['cd'])
        res = extract_pattern('{path}/{to}/{file}.txt', ['/tmp//1.txt'])
        self.assertEqual(res['path'], [None])
        self.assertEqual(res['to'], [None])
        self.assertEqual(res['file'], [None])
        res = extract_pattern('{path}/{to}/{file}.txt', ['/tmp/test/1.txt.txt'])
        self.assertEqual(res['path'], ['/tmp'])
        self.assertEqual(res['to'], ['test'])
        self.assertEqual(res['file'], ['1.txt'])
        # expand_pattern
        env.sos_dict = WorkflowDict({
            'os': os,
            'a': 100,
            'b': 'file name',
            'c': ['file1', 'file2', 'file 3'],
            'd': {
                'a': 'file1',
                'b': 'file2'
            },
        })
        self.assertEqual(expand_pattern('{b}.txt'), ['file name.txt'])
        self.assertEqual(
            expand_pattern('{c}.txt'), ['file1.txt', 'file2.txt', 'file 3.txt'])
        self.assertEqual(
            expand_pattern('{a}_{c}.txt'),
            ['100_file1.txt', '100_file2.txt', '100_file 3.txt'])

    def testAccessedVars(self):
        '''Test accessed vars of a SoS expression or statement.'''
        self.assertEqual(accessed_vars('''a = 1'''), set())
        self.assertEqual(accessed_vars('''a = b + 2.0'''), {'b'})
        self.assertEqual(accessed_vars('''a = "C"'''), set())
        self.assertEqual(accessed_vars('''a = "C" + f"{D}"'''), {'D'})
        self.assertEqual(accessed_vars('''a = 1 + f"{D + 20:f}" '''), {'D'})
        self.assertEqual(
            accessed_vars(
                '''k, "a.txt", "b.txt", par=f(something) ''', mode='eva'),
            {'k', 'f', 'something'})
        # this is a complicated case because the actual variable depends on the
        # result of an expression... However, in the NO-evaluation case, this is
        # the best we can do.
        self.assertEqual(accessed_vars('''c + f"{D + 1}" '''), {'c', 'D'})
        self.assertEqual(
            accessed_vars('''a, b=2, c=d ''', mode='eva'), {'a', 'd'})

    def testProgressBar(self):
        '''Test progress bar'''
        env.verbosity = 1
        #
        script = SoS_Script('''

[1]
[2]
[3]
[4]
[5]
''')
        wf = script.workflow()
        Base_Executor(wf).run()

    def testTextRepr(self):
        # the " as the last character can lead to problems...
        script = SoS_Script('''
run:
    echo "Hi, This is from bash"''')
        wf = script.workflow()
        Base_Executor(wf).run()
        # windows does not have #! mechanism so the python code
        # would incorrectly be executed as bat
        if sys.platform == 'win32':
            return
        for text in ('"""a"""', '"b"', r'"""\na\\nb"""', r"'''a\nb'''",
                     """ "a'\\"='" """):
            script = SoS_Script(r'''
a = 1
run: expand=True
   #!/usr/bin/env python
   with open('tmp.txt', 'w') as tmp:
      tmp.write({} + '{}')
k = """b"""'''.format(text, '{a}'))
            wf = script.workflow()
            Base_Executor(wf).run()
            with open('tmp.txt') as tmp:
                self.assertEqual(tmp.read(), eval(text) + '1')
        os.remove('tmp.txt')

    def testAnalyzeSection(self):
        '''Test analysis of sections (statically)'''
        script = SoS_Script('''
g1 = 'a'
g2 = 1
parameter: p1 = 5
parameter: infiles = 'a.txt'

[A_1: shared='b']
b = p1 + 2
input:  infiles
output: None
c = 5

[A_2]
b = [1, 2, 3]
input: for_each='b'
depends: 'some.txt', executable('ls')
import time
import random

r = random.randint(1, 5)
time.sleep(r)

[A_3]
input: None
print(p1)

[A_4]
input: None
task:
python: expand=True
   print(f'{output}')

[A_5]
task:
   print(f'{_output}')
''')
        wf = script.workflow('A')
        Base_Executor(wf)
        for section in wf.sections:
            res = analyze_section(section)
            if section.names[0][1] == '1':
                self.assertTrue(res['step_input'].undetermined())
                self.assertEqual(res['step_depends'], sos_targets())
                self.assertEqual(res['step_output'], sos_targets())
                self.assertEqual(res['environ_vars'], {'b', 'p1', 'infiles'})
                self.assertEqual(res['signature_vars'], {'c'})
                self.assertEqual(res['changed_vars'], {'b'})
            elif section.names[0][1] == '2':
                self.assertEqual(res['step_input'], sos_targets())
                self.assertEqual(res['step_depends'],
                                 sos_targets('some.txt', executable('ls')))
                self.assertTrue(res['step_output'].unspecified())
                # for_each will not be used for DAG
                self.assertEqual(res['environ_vars'],
                                 {'b', 'for_each', 'executable'})
                self.assertEqual(res['signature_vars'], {'r', 'time', 'random'})
                self.assertEqual(res['changed_vars'], set())
            elif section.names[0][1] == '4':
                self.assertTrue('output' in res['signature_vars'])
            elif section.names[0][1] == '5':
                self.assertTrue('output' not in res['signature_vars'])

    def testAnalyzeOutputFrom(self):
        '''Test extracting of from=value option from input'''
        script = SoS_Script('''
[A_1]
input:  output_from('B')

[A_2]
input: something_unknown, sos_groups(output_from(['C1', 'C2']), by=2), group_by=1
''')
        wf = script.workflow('A')
        Base_Executor(wf)
        for section in wf.sections:
            res = analyze_section(section)
            if section.names[0][1] == 1:
                self.assertEqual(res['step_depends'],
                                 sos_targets(sos_step('B')))
            if section.names[0][1] == 2:
                self.assertTrue(res['step_depends'] == sos_targets(
                    sos_step('C1'), sos_step('C2')))

    def testOnDemandOptions(self):
        '''Test options that are evaluated upon request.'''
        options = on_demand_options({'a': '"est"', 'b': 'c', 'c': 'e + 2'})
        env.sos_dict = WorkflowDict({
            'e': 10,
        })
        self.assertEqual(options['a'], 'est')
        self.assertRaises(KeyError, options.__getitem__, 'd')
        self.assertRaises(ValueError, options.__getitem__, 'b')
        self.assertEqual(options['c'], 12)
        #
        env.sos_dict.set('e', 20)
        self.assertEqual(options['c'], 22)
        env.sos_dict.set('c', 200)
        self.assertEqual(options['b'], 200)

    def testStableRepr(self):
        self.assertEqual(stable_repr({1, 2, '3', '1'}), "{'1', '3', 1, 2}")
        self.assertEqual(stable_repr({1: 2, 3: 4}), "{1:2, 3:4}")
        self.assertEqual(stable_repr([1, 3, 4]), "[1, 3, 4]")

    def testFileSig(self):
        '''test save and validate of file signature'''
        with open('test_sig.txt', 'w') as ts:
            ts.write('ba')
        a = file_target('test_sig.txt')
        a.write_sig()
        self.assertTrue(a.validate())
        #
        a.zap()
        self.assertTrue(a.validate())
        with open('test_sig.txt', 'w') as ts:
            ts.write('bac')
        self.assertFalse(a.validate())

    def testSplitFstring(self):
        '''Test function to split f-string in pieces '''
        for string, pieces in [
            ('hello world', ['hello world']),
            ('hello world {', None),
            ('{ hello world', None),
            ('hello world {}', None),
            ('hello world {a b}', None),
            ('{{hello world', ['{{hello world']),
            ('hello {{}} world', ['hello {{}} world']),
            ('hello {{ world }}', ['hello {{ world }}']),
            ('hello }} world', ['hello }} world']),
            ('hello {1} world', ['hello ', '1', ' world']),
            ('hello {a+b } }} world', ['hello ', 'a+b ', ' }} world']),
            ('hello {a+b:r} }} world', ['hello ', 'a+b:r', ' }} world']),
            ('hello {{{a+b!r} }} world', ['hello {{', 'a+b!r', ' }} world']),
            ('hello {a+b + {1,2}.pop() } }} world',
             ['hello ', 'a+b + {1,2}.pop() ', ' }} world']),
        ]:
            if pieces is None:
                self.assertRaises(SyntaxError, split_fstring, string)
            else:
                self.assertEqual(split_fstring(string), pieces)

    def testAsFstring(self):
        '''Test as fstring '''
        for string, fstring in [
            ('hello world', 'fr"""hello world"""'),
            ('{{hello world', 'fr"""{{hello world"""'),
            ('hello {{}} world', 'fr"""hello {{}} world"""'),
            ('hello {{ world }}', 'fr"""hello {{ world }}"""'),
            ('hello }} \n world', 'fr"""hello }} \n world"""'),
            (r'hello }} \n world', 'fr"""hello }} \\n world"""'),
            ('hello {1} world', 'fr"""hello {1} world"""'),
            ('hello {a+b } }} world', 'fr"""hello {a+b } }} world"""'),
            ('hello {a+b:r} }} world', 'fr"""hello {a+b:r} }} world"""'),
            ('''hello """ \'\'\' {a+b } }} world''',
             'f\'hello """ \\\'\\\'\\\' {a+b } }} world\''),
            ('hello {{{a+b!r} }} world', 'fr"""hello {{{a+b!r} }} world"""'),
            ('hello {a+b + {1,2}.pop() } }} world',
             'fr"""hello {a+b + {1,2}.pop() } }} world"""'),
            ('''hello {'a'+b !r} }} world''',
             'fr"""hello {\'a\'+b !r} }} world"""'),
            ('''hello """ \'\'\' {'a'+"b" + {"c", "D"}.pop() } }} world''',
             '\'hello """ \\\'\\\'\\\' {0} }} world\'.format(\'a\'+"b" + {"c", "D"}.pop() )'
            ),
        ]:
            self.assertEqual(as_fstring(string), fstring)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
#
# Copyright (c) Bo Peng and the University of Texas MD Anderson Cancer Center
# Distributed under the terms of the 3-clause BSD License.

import os
import shutil
import subprocess
import unittest
from pathlib import Path


class TestRemove(unittest.TestCase):

    def setUp(self):
        if os.path.isdir('temp'):
            shutil.rmtree('temp')
        os.mkdir('temp')
        os.chdir('temp')
        with open('test.sos', 'w') as script:
            script.write('''
[0]
output:  't_f1'
run:
    touch t_f1

[1]
output:  't_d1/t_f2'
run:
    touch t_d1/t_f2
    touch t_d1/ut_f4

[2]
output:  't_d2/t_d3/t_f3'
run:
    touch t_d2/t_d3/t_f3

''')
        subprocess.call('sos run test -s force', shell=True)
        # create some other files and directory
        for d in ('ut_d1', 'ut_d2', 'ut_d2/ut_d3'):
            os.mkdir(d)
        for f in ('ut_f1', 'ut_d1/ut_f2', 'ut_d2/ut_d3/ut_f3'):
            with open(f, 'w') as tf:
                tf.write(f)

    def assertExists(self, fdlist):
        for fd in fdlist:
            self.assertTrue(os.path.exists(fd), '{} does not exist'.format(fd))

    def assertNonExists(self, fdlist):
        for fd in fdlist:
            self.assertFalse(os.path.exists(fd), '{} still exists'.format(fd))

    def testSetup(self):
        self.assertExists([
            'ut_d1', 'ut_d2', 'ut_d2/ut_d3', 'ut_f1', 'ut_d1/ut_f2',
            'ut_d2/ut_d3/ut_f3'
        ])
        self.assertExists(
            ['t_f1', 't_d1/t_f2', 't_d2/t_d3/t_f3', 't_d2/t_d3', 't_d2'])
        # this is the tricky part, directory containing untracked file should remain
        self.assertExists(['t_d1', 't_d1/ut_f4'])

    def testRemoveAllTracked(self):
        '''test list files'''
        subprocess.call('sos remove . -t -y', shell=True)
        self.assertExists([
            'ut_d1', 'ut_d2', 'ut_d2/ut_d3', 'ut_f1', 'ut_d1/ut_f2',
            'ut_d2/ut_d3/ut_f3'
        ])
        self.assertNonExists(['t_d1/t_f2', 't_d2/t_d3/t_f3'])
        # this is the tricky part, directory containing untracked file should remain
        self.assertExists(['t_d1', 't_f1', 't_d1/ut_f4'])

    def testRemoveSpecificTracked(self):
        # note the t_f1, which is under current directory and has to be remove specifically.
        subprocess.call('sos remove t_f1 ut_f1 t_d2 ut_d2 -t -y', shell=True)
        self.assertExists([
            'ut_d1', 'ut_d2', 'ut_d2/ut_d3', 'ut_f1', 'ut_d1/ut_f2',
            'ut_d2/ut_d3/ut_f3', 't_d1/t_f2', 't_d1', 't_d1/ut_f4'
        ])
        self.assertNonExists(['t_f1', 't_d2/t_d3/t_f3'])

    def testRemoveAllUntracked(self):
        '''test remove all untracked files'''
        subprocess.call('sos remove . -u -y', shell=True)
        self.assertNonExists(['ut_d1/ut_f2', 't_d1/ut_f4', 'ut_d2/ut_d3/ut_f3'])
        self.assertExists([
            't_d1/t_f2', 't_d2/t_d3/t_f3', 't_d2/t_d3', 't_d2', 't_d1', 't_f1'
        ])
        # this is the tricky part, files under the current directory are not removed
        self.assertExists(['ut_f1'])

    def testRemoveSpecificUntracked(self):
        # note the t_f1, which is under current directory and has to be remove specifically.
        subprocess.call(
            'sos remove t_f1 ut_f1 ut_d1/ut_f2 t_d1 -u -y', shell=True)
        self.assertNonExists(['ut_f1', 'ut_d1/ut_f2', 't_d1/ut_f4'])
        self.assertExists([
            't_d1/t_f2', 't_d2/t_d3/t_f3', 't_d2/t_d3', 't_d2', 't_d1', 't_f1'
        ])
        self.assertExists(
            ['ut_d1', 'ut_d2', 'ut_d2/ut_d3', 'ut_d2/ut_d3/ut_f3'])

    def testRemoveByAge(self):
        '''test remove by age'''
        subprocess.call('sos remove --age=+1h -y', shell=True)
        # nothing is removed
        self.assertExists([
            'ut_f1', 'ut_d1/ut_f2', 't_d1/ut_f4', 't_d1/t_f2', 't_d2/t_d3/t_f3',
            't_d2/t_d3', 't_d2', 't_d1', 't_f1', 'ut_d1', 'ut_d2',
            'ut_d2/ut_d3', 'ut_d2/ut_d3/ut_f3'
        ])
        #
        subprocess.call('sos remove -t --age=-1h -y', shell=True)
        self.assertNonExists(['t_d1/t_f2', 't_d2/t_d3/t_f3'])
        self.assertExists([
            'ut_f1', 'ut_d1/ut_f2', 't_d1/ut_f4', 't_f1', 't_d2/t_d3', 't_d2',
            't_d1', 'ut_d1', 'ut_d2', 'ut_d2/ut_d3', 'ut_d2/ut_d3/ut_f3'
        ])
        #
        subprocess.call('sos remove -u --age=-1h -y', shell=True)
        self.assertExists([
            'ut_f1', 't_f1', 't_d2/t_d3', 't_d2', 't_d1', 'ut_d1', 'ut_d2',
            'ut_d2/ut_d3'
        ])

    def testRemoveBySize(self):
        '''test remove by size'''
        subprocess.call('sos remove --size=+10M -y', shell=True)
        # nothing is removed
        self.assertExists([
            'ut_f1', 'ut_d1/ut_f2', 't_d1/ut_f4', 't_d1/t_f2', 't_d2/t_d3/t_f3',
            't_d2/t_d3', 't_d2', 't_d1', 't_f1', 'ut_d1', 'ut_d2',
            'ut_d2/ut_d3', 'ut_d2/ut_d3/ut_f3'
        ])
        #
        subprocess.call('sos remove -t --size=-1M -y', shell=True)
        self.assertNonExists(['t_d1/t_f2', 't_d2/t_d3/t_f3'])
        self.assertExists([
            'ut_f1', 'ut_d1/ut_f2', 't_d1/ut_f4', 't_f1', 't_d2/t_d3', 't_d2',
            't_d1', 'ut_d1', 'ut_d2', 'ut_d2/ut_d3', 'ut_d2/ut_d3/ut_f3'
        ])
        #
        subprocess.call('sos remove -u --size=-1M -y', shell=True)
        self.assertExists([
            'ut_f1', 't_f1', 't_d2/t_d3', 't_d2', 't_d1', 'ut_d1', 'ut_d2',
            'ut_d2/ut_d3'
        ])

    def testRemoveAll(self):
        '''Test remove all specified files'''
        subprocess.call('sos remove ut_d1 t_d1 ut_d2/ut_d3 -y', shell=True)
        self.assertExists(['t_d2/t_d3/t_f3', 't_d2/t_d3', 't_d2', 't_f1'])

    def testRemovePlaceholders(self):
        '''Test remaining placeholder files'''
        # let us create a fake placeholder file
        os.remove('t_f1')
        os.remove('t_d1/t_f2')
        subprocess.call('sos dryrun test.sos', shell=True)
        self.assertFalse(os.path.isfile('t_f1'))
        self.assertFalse(os.path.isfile('t_d1/t_f2'))
        Path('t_f1').touch()
        Path('t_d1/t_f2').touch()
        #
        subprocess.call('sos remove -p', shell=True)
        self.assertFalse(os.path.isfile('t_f1'))
        self.assertFalse(os.path.isfile('t_d1/t_f2'))

    def tearDown(self):
        os.chdir('..')
        shutil.rmtree('temp')


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
#
# Copyright (c) Bo Peng and the University of Texas MD Anderson Cancer Center
# Distributed under the terms of the 3-clause BSD License.

import copy
import glob
import os
import pickle
import re
import shlex
import shutil
import subprocess
import sys
from collections.abc import Iterable, Sequence
from copy import deepcopy
from itertools import combinations, tee
from pathlib import Path, WindowsPath, PosixPath
from shlex import quote
from typing import Union, Dict, Any, List
import fasteners
import pkg_resources

from .utils import (Error, env, pickleable, short_repr, stable_repr)
from .pattern import extract_pattern
from .eval import interpolate
from .controller import request_answer_from_controller, send_message_to_controller

try:
    from xxhash import xxh64 as hash_md5
except ImportError:
    from hashlib import md5 as hash_md5

__all__ = ['dynamic', 'executable', 'env_variable', 'sos_variable']


def is_basic_type(obj):
    if isinstance(obj, (bool, int, float, str, bytes)):
        return True
    if isinstance(obj, (tuple, list, set)):
        return all(is_basic_type(x) for x in obj)
    if isinstance(obj, dict):
        return all(is_basic_type(x) for x in obj.keys()) and all(
            is_basic_type(x) for x in obj.values())
    if isinstance(obj, (file_target, path, paths)):
        return True
    # we support types defined in numpy and pandas, but not others
    module = obj.__class__.__module__
    if 'pandas' in module or 'numpy' in module:
        return True
    return False


class UnknownTarget(Error):

    def __init__(self, target: 'BaseTarget'):
        Error.__init__(self, 'Target unavailable: %s' % target)
        self.target = target


class RemovedTarget(Error):

    def __init__(self, target: 'BaseTarget'):
        if isinstance(target, invalid_target):
            Error.__init__(self, 'Invalid target from errored steps')
        else:
            Error.__init__(self, 'Target removed: %s' % target)
        self.target = target


class UnavailableLock(Error):
    """Raised when there are errors in prepare mode. Such errors are not raised
    immediately, but will be collected and raised at the end """

    def __init__(self, signature):
        Error.__init__(
            self,
            f'Failed to obtain lock {signature[2]} for input {short_repr(signature[0])} and output {short_repr(signature[1])}. It is likely '
            +
            'that these files are protected by another SoS process or concurrant task that is generating the same set of files. Please manually remove the lockfile '
            + 'if you are certain that no other process is using the lock.')
        self.input = signature[0]
        self.output = signature[1]
        self.sig_file = signature[2]


#
# Runtime signature
#


def textMD5(text):
    '''Get md5 of a piece of text'''
    m = hash_md5()
    if isinstance(text, str):
        m.update(text.encode())
    else:
        m.update(text)
    return m.hexdigest()


def objectMD5(obj):
    '''Get md5 of an object'''
    if hasattr(obj, 'target_name'):
        return obj.target_name()
    try:
        return textMD5(pickle.dumps(obj))
    except Exception:
        return ''


def fileMD5(filename, partial=True):
    '''Calculate partial MD5, basically the first and last 8M
    of the file for large files. This should signicicantly reduce
    the time spent on the creation and comparison of file signature
    when dealing with large bioinformat ics datasets. '''
    filesize = os.path.getsize(filename)
    # calculate md5 for specified file
    md5 = hash_md5()
    block_size = 2**20  # buffer of 1M
    try:
        # 2**24 = 16M
        if (not partial) or filesize < 2**24:
            with open(filename, 'rb') as f:
                while True:
                    data = f.read(block_size)
                    if not data:
                        break
                    md5.update(data)
        else:
            count = 16
            # otherwise, use the first and last 8M
            with open(filename, 'rb') as f:
                while True:
                    data = f.read(block_size)
                    count -= 1
                    if count == 8:
                        # 2**23 = 8M
                        f.seek(-2**23, 2)
                    if not data or count == 0:
                        break
                    md5.update(data)
    except IOError as e:
        sys.exit(f'Failed to read {filename}: {e}')
    return md5.hexdigest()


class BaseTarget(object):
    '''A base class for all targets (e.g. a file)'''

    def __init__(self, *args, **kwargs):
        self._sigfile = None
        self._dict = kwargs
        self.traced = False

    def set_traced(self):
        self.traced = True

    def set(self, *args, **kwargs):
        if args:
            if len(args) != 2:
                raise ValueError(
                    'set(name, value) or set(name=value) is expected.')
            if not is_basic_type(args[1]):
                env.logger.warning(
                    f'Failed to set attribute: {args[1]} is or contains unsupported data type.'
                )
                return self
            if hasattr(self, args[0]):
                raise ValueError(
                    f'Attribute {args[0]} conflicts with another attribute of type {self.__class__.__name__}.'
                )
            self._dict[args[0]] = args[1]
        #
        if kwargs:
            for name, value in kwargs.items():
                if not is_basic_type(value):
                    env.logger.warning(
                        f'Failed to set attribute: {value} is or contains unsupported data type.'
                    )
                    return self
                if hasattr(self, name):
                    raise ValueError(
                        f'Attribute {name} conflicts with another attribute of {value.__class__.__name__}.'
                    )
            self._dict.update(kwargs)
        return self

    def get(self, name, default=None):
        return self._dict.get(name, default)

    def __getattr__(self, name):
        try:
            return self._dict[name]
        except Exception:
            # if name in self._dict:
            # return self._dict.get(name)
            raise AttributeError(
                f'{self.__class__.__name__} object has no attribute {name}')

    def target_exists(self, mode='any'):
        # mode should be 'any', 'target', or 'signature'
        raise RuntimeError('Undefined base function target_exists')

    def target_name(self):
        # name of the target, which should be able to differentiate
        # this object with other targets of the same type.
        raise RuntimeError('Undefined base function target_name')

    def target_signature(self):
        # signature of the content of the target, which should be
        # able to detect changes of the content of target
        raise RuntimeError('Undefined base function target_signature')

    def create_placeholder(self):
        pass

    def __repr__(self):
        return f'{self.__class__.__name__}("{self.target_name()}")'

    def __hash__(self):
        return hash(repr(self))

    def validate(self, sig):
        return self.target_signature() == sig

    def __eq__(self, obj):
        return isinstance(obj, self.__class__) and self.target_signature(
        ) == obj.target_signature()


class sos_variable(BaseTarget):
    '''A target for a SoS variable.'''

    def __init__(self, var):
        super(sos_variable, self).__init__()
        self._var = var

    def target_exists(self, mode='any'):
        return self._var in env.sos_dict

    def target_name(self):
        return self._var

    def target_signature(self):
        return objectMD5(env.sos_dict[self._var])

    def __hash__(self):
        return hash(repr(self))

    def __eq__(self, other):
        return isinstance(other, sos_variable) and self._var == other._var

    def __format__(self, format_spec):
        # handling special !q conversion flag
        if format_spec and format_spec[0] == 'R':
            return self._var.__format__(format_spec[1:])
        else:
            return str(self).__format__(format_spec)


class env_variable(BaseTarget):
    '''A target for an environmental variable.'''

    def __init__(self, var):
        super(env_variable, self).__init__()
        self._var = var

    def target_exists(self, mode='any'):
        return self._var in os.environ

    def target_name(self):
        return self._var

    def target_signature(self, mode='any'):
        return textMD5(repr(os.environ[self._var]))

    def __eq__(self, other):
        return isinstance(other, env_variable) and self._var == other._var

    def __hash__(self):
        return hash(repr(self))

    def __format__(self, format_spec):
        # handling special !q conversion flag
        if format_spec and format_spec[0] == 'R':
            return self._var.__format__(format_spec[1:])
        else:
            return str(self).__format__(format_spec)


class invalid_target(BaseTarget):
    '''A target for an environmental variable.'''

    def __init__(self):
        super(invalid_target, self).__init__()

    def target_exists(self, mode='any'):
        return False

    def target_name(self):
        return 'invalid'

    def target_signature(self, mode='any'):
        return ''


class system_resource(BaseTarget):
    '''A target for required computing resource.'''

    def __init__(self, mem=None, disk=None):
        super(system_resource, self).__init__()
        self._mem = mem
        self._disk = disk

    def target_exists(self, mode='any'):
        if self._mem:
            import psutil
            from .utils import expand_size
            avail = psutil.virtual_memory().available
            if avail < expand_size(self._mem):
                #env.logger.warning(f'System available memory {avail} is less than request {self._mem}')
                return False
        if self._disk:
            import psutil
            from .utils import expand_size
            avail = psutil.disk_usage(os.path.abspath('.')).free
            if avail < expand_size(self._disk):
                #env.logger.warning(f'System available diskspace {avail} is less than request {self._disk}')
                return False
        return True

    def target_name(self):
        res = []
        if self._mem:
            res.append(f'mem={repr(self._mem)}')
        if self._disk:
            res.append(f'disk={repr(self._disk)}')
        return f'system_resource({",".join(res)})'

    def target_signature(self):
        return ''

    def __repr__(self):
        return self.target_name()


class sos_step(BaseTarget):
    '''A target for a step of sos.'''

    def __init__(self, step_name, **kwargs):
        super(sos_step, self).__init__(**kwargs)
        self._step_name = str(step_name)

    def target_exists(self, mode='any'):
        # the target exists only if it has been executed?
        # which is indicated by a variable
        return request_answer_from_controller(['sos_step', self._step_name])

    def target_name(self):
        return self._step_name

    def target_signature(self, mode='any'):
        return textMD5(f'sos_step({self._step_name})')

    def __eq__(self, other):
        return isinstance(other,
                          sos_step) and self._step_name == other._step_name

    def __hash__(self):
        return hash(repr(self))

    def __format__(self, format_spec):
        # handling special !q conversion flag
        if format_spec and format_spec[0] == 'R':
            return self._step_name.__format__(format_spec[1:])
        return str(self).__format__(format_spec)


class named_output(BaseTarget):
    '''A target for a named output'''

    def __init__(self, output_name):
        super(named_output, self).__init__()
        if not isinstance(output_name, str):
            raise ValueError('named_output() only accept one output name')
        self._output_name = output_name

    def target_exists(self, mode='any'):
        # this target is handled specifically and will not be used directly
        return False

    def target_name(self):
        return self._output_name

    def target_signature(self, mode='any'):
        return textMD5(f'named_output({self._output_name})')

    def __eq__(self, other):
        return isinstance(
            other, named_output) and self._output_name == other._output_name

    def __hash__(self):
        return hash(repr(self))


class dynamic(BaseTarget):
    '''A dynamic executable that only handles input files when
    it is available. This target is handled directly with its `resolve`
    function called by the executor. '''

    def __init__(self, target):
        self._target = target

    def target_name(self):
        return self._target

    def target_exists(self, mode='any'):
        return False

    def target_signature(self, mode='any'):
        return ''

    def resolve(self):
        if isinstance(self._target, str):
            return sorted(
                [x for x in glob.glob(self._target) if os.path.isfile(x)])
        elif isinstance(self._target, Sequence) and all(
                isinstance(x, str) for x in self._target):
            return sorted(
                sum([[x
                      for x in glob.glob(t)
                      if os.path.isfile(x)]
                     for t in self._target], []))
        else:
            return self._target

    def __format__(self, format_spec):
        # handling special !q conversion flag
        if format_spec and format_spec[0] == 'R':
            return sos_targets(self._target).__format__(format_spec[1:])
        else:
            return str(self).__format__(format_spec)


class remote(BaseTarget):
    '''A remote target is not tracked and not translated during task execution'''

    def __init__(self, *targets):
        super(remote, self).__init__()
        self.__unresolvable_object__ = True
        if len(targets) == 1:
            self._target = targets[0]
        else:
            # multi-item targets
            self._target = targets
        if isinstance(self._target,
                      Sequence) and not isinstance(self._target, str):
            self.__flattenable__ = True

    def target_name(self):
        if isinstance(self._target, str):
            return file_target(self._target).target_name()
        elif isinstance(self._target, BaseTarget):
            return self._target.target_name()
        else:
            return repr(self._target)

    def target_exists(self, mode='any'):
        return True

    def target_signature(self, mode='any'):
        return textMD5(self.target_name())

    def resolve(self):
        return self._target

    def flatten(self):
        return [remote(x) for x in self._target]

    def __format__(self, format_spec):
        # handling special !q conversion flag
        if format_spec and format_spec[0] == 'R':
            return sos_targets(self._target).__format__(format_spec[1:])
        else:
            return str(self).__format__(format_spec)


class executable(BaseTarget):
    '''A target for an executable command.'''

    def __init__(self, cmd, version=None):
        super(executable, self).__init__()
        self._cmd = cmd
        self._md5 = None
        if version is None:
            self._version = ()
        elif isinstance(version, str):
            self._version = (version,)
        else:
            self._version = tuple(version)

    def __eq__(self, other):
        return isinstance(
            other, executable
        ) and self._cmd == other._cmd and self._version == other._version

    def target_exists(self, mode='any'):
        if mode in ('any', 'target') and shutil.which(
                shlex.split(self._cmd)[0]):
            if self._version:
                try:
                    output = subprocess.check_output(
                        self._cmd,
                        stderr=subprocess.STDOUT,
                        shell=True,
                        timeout=5).decode()
                except subprocess.TimeoutExpired as e:
                    env.logger.warning(str(e))
                    return False
                except subprocess.CalledProcessError as e:
                    env.logger.warning(str(e))
                    return False
                for ver in self._version:
                    if ver in output:
                        return True
                return False
            else:
                return True
        return False

    def target_name(self):
        if self._version:
            return f'{self._cmd} (version={self._version})'
        else:
            return self._cmd

    def target_signature(self):
        # we do not care if the target actually exist
        # or its content has changed
        return textMD5(self._cmd)

    def __hash__(self):
        return hash(repr(self))

    def __format__(self, format_spec):
        # handling special !q conversion flag
        if format_spec and format_spec[0] == 'R':
            return self._cmd.__format__(format_spec[1:])
        else:
            return str(self).__format__(format_spec)


def collapseuser(path):
    home = os.path.expanduser('~')
    if path == home:
        return '~'
    elif path.startswith(home + os.sep):
        return '~' + path[len(home):]
    else:
        return path


class path(type(Path())):
    '''A regular target for files.
    '''
    CONVERTERS = {
        'u':
            os.path.expanduser,
        'U':
            collapseuser,
        'e':
            lambda x: x.replace(' ', '\\ '),
        'a':
            lambda x: os.path.abspath(x),
        'l':
            lambda x: os.path.realpath(x),
        'd':
            lambda x: os.path.dirname(x) or '.',
        'b':
            os.path.basename,
        'n':
            lambda x: os.path.splitext(x)[0],
        'x':
            lambda x: os.path.splitext(x)[1],
        'q': (lambda x: subprocess.list2cmdline([x]))
             if sys.platform == 'win32' else quote,
        'p':
            lambda x: ('/' if len(x) > 1 and x[1] == ':' else '') + x.replace(
                '\\', '/').replace(':/', '/').replace(':', '/'),
        'r':
            repr,
        's':
            str,
        # these are handled elsewhere
        ',':
            lambda x: x,
        '!':
            lambda x: x,
        'R':
            lambda x: x,
    }

    def __new__(cls, *args, **kwargs):
        if cls is Path:
            cls = WindowsPath if os.name == 'nt' else PosixPath

        if 'name' in kwargs:
            try:
                if 'default' in kwargs:
                    return cls._from_parts([
                        env.sos_dict['CONFIG']['hosts'][
                            env.sos_dict['__host__']]['paths'].get(
                                kwargs['name'], kwargs['default'])
                    ]).expanduser()
                else:
                    return cls._from_parts([
                        env.sos_dict['CONFIG']['hosts'][
                            env.sos_dict['__host__']]['paths'][kwargs['name']]
                    ]).expanduser()
            except Exception:
                if '__host__' not in env.sos_dict:
                    raise RuntimeError(
                        'Incomplete sos environment: missing __host__ definition.'
                    )
                if 'CONFIG' not in env.sos_dict or 'hosts' not in env.sos_dict[
                        'CONFIG']:
                    raise RuntimeError(
                        'Incomplete sos environment: missing hosts definition.')
                if env.sos_dict['__host__'] not in env.sos_dict['CONFIG'][
                        'hosts']:
                    raise RuntimeError(
                        f'Incomplete sos environment: undefined host {env.sos_dict["__host__"]}'
                    )
                if 'paths' not in env.sos_dict['CONFIG']['hosts'][
                        env.sos_dict['__host__']]:
                    raise RuntimeError(
                        f'Incomplete sos environment: paths not defined for host {env.sos_dict["__host__"]}'
                    )
                if kwargs['name'] not in env.sos_dict['CONFIG']['hosts'][
                        env.sos_dict['__host__']]['paths']:
                    raise ValueError(
                        f'{kwargs["name"]} not defined for host {env.sos_dict["__host__"]}'
                    )
        return cls._from_parts(args).expanduser()

    @staticmethod
    def names(host=None):
        if host is None:
            if '__host__' not in env.sos_dict:
                raise RuntimeError(
                    'Incomplete sos environment: missing __host__ definition.')
            host = env.sos_dict['__host__']
        if 'CONFIG' not in env.sos_dict or 'hosts' not in env.sos_dict['CONFIG']:
            raise RuntimeError(
                'Incomplete sos environment: missing hosts definition.')
        if host not in env.sos_dict['CONFIG']['hosts']:
            raise RuntimeError(
                'Incomplete sos environment: undefined host {host}')
        if 'paths' not in env.sos_dict['CONFIG']['hosts'][host]:
            return []
        else:
            return list(env.sos_dict['CONFIG']['hosts'][host]['paths'].keys())

    def _init(self, template=None):
        super(path, self)._init(template)
        if not (self._drv or
                self._root) and self._parts and self._parts[0][:1] == '~':
            expanded = self.expanduser()
            self._parts = expanded._parts
            self._drv = expanded._drv
            self._root = expanded._root

    def is_external(self):
        try:
            return os.path.relpath(self.fullname(),
                                   env.exec_dir).startswith('..')
        except Exception:
            # under windows the file might be on different volume
            return True

    def fullname(self):
        return os.path.abspath(str(self))

    def __fspath__(self):
        return self.fullname()

    def __eq__(self, other):
        return os.path.abspath(self.fullname()) == os.path.abspath(
            (other
             if isinstance(other, file_target) else path(other)).fullname())

    def __add__(self, part):
        if isinstance(part, (str, path)):
            return self.__class__(str(self) + str(part))
        else:
            raise ValueError(
                f"Cannot concatenate path to {part} of type {type(part).__name__}: expect a string or path"
            )

    def __format__(self, format_spec):
        # handling special !q conversion flag
        obj = str(self)
        for i, c in enumerate(format_spec):
            if c in self.CONVERTERS:
                obj = self.CONVERTERS[c](obj)
            else:
                # other defined format
                return obj.__format__(format_spec[i:])
        return obj

    def __lt__(self, other):
        return str(self) < str(other)

    def __hash__(self):
        return hash(repr(self))

    def zap(self):
        zap_file = self + '.zapped'
        if not self.exists() and zap_file.is_file():
            return
        if not self.exists() or not self.is_file():
            raise FileNotFoundError(str(self))
        with open(zap_file, 'w') as md5:
            md5.write(
                f'{self.resolve()}\t{os.path.getmtime(self)}\t{os.path.getsize(self)}\t{fileMD5(self)}\n'
            )
        self.unlink()


class file_target(path, BaseTarget):
    '''A regular target for files.
    '''

    def __init__(self, *args, **kwargs):
        # this is path segments
        super(file_target, self).__init__(*args, **kwargs)
        if len(args) == 1 and isinstance(args[0], file_target):
            self._md5 = args[0]._md5
        else:
            self._md5 = None

    def _init(self, template=None):
        super(file_target, self)._init(template)
        self._md5 = None

    def create_placeholder(self):
        # create an empty placeholder file
        if 'TARGET' in env.config['SOS_DEBUG'] or 'ALL' in env.config[
                'SOS_DEBUG']:
            env.log_to_file('TARGET', f'Create placeholder target {self}')
        self.touch()
        send_message_to_controller(
            ['workflow_sig', 'placeholder', 'file_target',
             str(self)])

    def target_exists(self, mode='any'):
        try:
            if mode in ('any', 'target') and self.exists():
                return True
            elif mode == 'any' and (self + '.zapped').exists():
                return True
            return False
        except Exception as e:
            env.logger.debug(f"Invalid file_target {self}: {e}")
            return False

    def size(self):
        try:
            return os.path.getsize(self)
        except Exception:
            if (self + '.zapped').is_file():
                with open(self + '.zapped') as sig:
                    line = sig.readline()
                    _, _, s, _ = line.strip().rsplit('\t', 3)
                    return int(s)

    def target_name(self):
        return str(self)

    def __fspath__(self):
        return super(file_target, self).__fspath__()

    def target_signature(self):
        '''Return file signature'''
        if self.exists():
            if not self._md5:
                self._md5 = fileMD5(self)
            return (os.path.getmtime(self), os.path.getsize(self), self._md5)
        elif (self + '.zapped').is_file():
            with open(self + '.zapped') as sig:
                line = sig.readline()
                _, mtime, size, md5 = line.strip().rsplit('\t', 3)
                self._md5 = md5
                return (float(mtime), int(size), md5)
        else:
            raise ValueError(f'{self} does not exist.')

    def sig_file(self):
        return os.path.join(env.exec_dir, '.sos',
                            f'{textMD5(str(self.resolve()))}.file_info')

    def validate(self, sig=None):
        '''Check if file matches its signature'''
        if sig is not None:
            sig_mtime, sig_size, sig_md5 = sig
        else:
            try:
                with open(self.sig_file()) as sig:
                    sig_mtime, sig_size, sig_md5 = sig.read().strip().split()
            except Exception:
                return False
        if not self.exists():
            if (self + '.zapped').is_file():
                with open(self + '.zapped') as sig:
                    line = sig.readline()
                    return sig_md5 == line.strip().rsplit('\t', 3)[-1]
            else:
                return False

        if int(sig_size) != os.path.getsize(self):
            return False
        if sig_mtime == os.path.getmtime(self):
            return True
        return fileMD5(self) == sig_md5

    def write_sig(self):
        '''Write signature to sig store'''
        if not self._md5:
            self._md5 = fileMD5(self)
        with open(self.sig_file(), 'w') as sig:
            sig.write(
                f'{os.path.getmtime(self)}\t{os.path.getsize(self)}\t{self._md5}'
            )

    def __hash__(self):
        return hash(repr(self))

    def __eq__(self, obj):
        return isinstance(
            obj, file_target) and os.path.abspath(self) == os.path.abspath(obj)

    def __deepcopy__(self, memo):
        ft = file_target(self)
        ft._dict = copy.deepcopy(self._dict)
        return ft

    def __reduce__(self):
        return tuple([
            self.__class__,
            super(file_target, self).__reduce__()[1], {
                '_md5': self._md5,
                '_dict': self._dict
            }
        ])


class paths(Sequence, os.PathLike):
    '''A collection of targets'''
    # check if string contains wildcard character
    wildcard = re.compile(r'[*?\[]')

    def __init__(self, *args):
        self._paths = []
        for arg in args:
            self.__append__(arg)

        for t in self._paths:
            if isinstance(t, paths):
                raise RuntimeError(
                    f"Nested paths {t} were introduced by {args}")
            if not isinstance(t, path):
                raise RuntimeError(f"Unrecognized path {t}")

    def paths(self):
        return [str(x) for x in self._paths]

    def __append__(self, arg):
        if isinstance(arg, paths):
            self._paths.extend(arg._paths)
        elif isinstance(arg, str):
            if self.wildcard.search(arg):
                matched = sorted(glob.glob(os.path.expanduser(arg)))
                if matched:
                    self._paths.extend([path(x) for x in matched])
                else:
                    env.logger.debug(f'Pattern {arg} does not match any file')
            else:
                self._paths.append(path(arg))
        elif isinstance(arg, sos_targets):
            if not all(isinstance(x, file_target) for x in arg._targets):
                raise ValueError(
                    f'Cannot convert a sos_targets object {arg} with non-file target to paths'
                )
            self._paths.extend([path(str(x)) for x in arg._targets])
        elif isinstance(arg, file_target):
            self._paths.append(path(str(arg)))
        elif isinstance(arg, Iterable):
            for t in list(arg):
                self.__append__(t)
        elif arg is not None:
            self._paths.append(path(str(arg)))

    def zap(self):
        for p in self._paths:
            p.zap()

    def __getstate__(self):
        return self._paths

    def __setstate__(self, paths):
        self._paths = paths

    def __len__(self):
        return len(self._paths)

    def __getitem__(self, i):
        return self._paths[i]

    def __fspath__(self):
        if len(self._paths) == 1:
            return self._paths[0].__fspath__()
        if not self._paths:
            raise ValueError(f"Cannot treat an empty paths as single path")
        raise ValueError(
            f'Cannot treat an paths object {self} with more than one paths as a single path'
        )

    def __format__(self, format_spec):
        if ',' in format_spec:
            fmt_spec = format_spec.replace(',', '')
            return ','.join(x.__format__(fmt_spec) for x in self._paths)
        return ' '.join(x.__format__(format_spec) for x in self._paths)

    def __deepcopy__(self, memo):
        return paths(deepcopy(self._paths))

    def __getattr__(self, key):
        if len(self._paths) == 1:
            return getattr(self._paths[0], key)
        if len(self._paths) == 0:
            raise AttributeError(
                f"Cannot get attribute {key} from empty target list")
        raise AttributeError(
            f'Cannot get attribute {key} from group of {len(self)} targets {self!r}'
        )

    def __hash__(self):
        return hash(repr(self))

    def __eq__(self, other):
        return self._paths == other._paths if isinstance(other,
                                                         paths) else other

    def __repr__(self):
        return '[' + ', '.join(repr(x) for x in self._paths) + ']'

    def __str__(self):
        return self.__format__('')


class _sos_group(BaseTarget):
    '''A type that is similar to sos_targets but saves index of objects '''

    def __init__(self, indexes, labels=None, parent=None):
        super(_sos_group, self).__init__()
        self._indexes = list(indexes)
        if labels is not None:
            if isinstance(labels, str):
                self._labels = [labels] * len(indexes)
            else:
                self._labels = labels
                if len(self._indexes) != len(self._labels):
                    raise ValueError('Index and source have different length')
        elif parent is not None:
            self._labels = [parent._labels[x] for x in indexes]
        else:
            raise ValueError('Either labels or indexes should be specified')

    def add_last(self, n, parent):
        # add the last n elements of parent to group
        # this has to be called after the elements have been appended
        self._indexes.extend(
            range(len(parent._targets) - n, len(parent._targets)))
        self._labels.extend(parent._labels[len(parent._targets) -
                                           n:len(parent._targets)])
        return self

    def extend(self, grp, start, parent):
        self._indexes.extend([x + start for x in grp._indexes])
        self._labels.extend([parent._labels[x + start] for x in grp._indexes])
        self._dict.update(grp._dict)
        return self

    def __repr__(self):
        return f'_sos_group(indexes={self._indexes}, labels={self._labels})'

    def idx_to_targets(self, parent):
        ret = sos_targets([])
        ret._targets = [parent._targets[x] for x in self._indexes]
        ret._labels = self._labels
        ret._dict = self._dict
        return ret

    def __getstate__(self):
        return dict(
            indexes=self._indexes, labels=self._labels, properties=self._dict)

    def __setstate__(self, sdict):
        self._indexes = sdict['indexes']
        self._labels = sdict['labels']
        self._dict = sdict['properties']


class sos_targets(BaseTarget, Sequence, os.PathLike):
    '''A collection of targets.
    If verify_existence is True, an UnknownTarget exception
    will be thrown if target does not exist.
    '''
    # check if string contains wildcard character
    wildcard = re.compile(r'[*?\[]')

    def __init__(self,
                 *args,
                 group_by=None,
                 paired_with=None,
                 pattern=None,
                 group_with=None,
                 for_each=None,
                 _undetermined: Union[bool, str] = None,
                 _source='',
                 _verify_existence=False,
                 **kwargs):
        super(sos_targets, self).__init__()
        self._targets: List = []
        self._labels: List = []
        self._groups: List = []
        if isinstance(_undetermined, (bool, str)):
            self._undetermined = _undetermined
        else:
            self._undetermined = not bool(args)
        for arg in args:
            self.__append__(
                arg, default_source=_source, verify_existence=_verify_existence)
        for src, value in kwargs.items():
            self.__append__(
                value, source=src, verify_existence=_verify_existence)
        for t in self._targets:
            if isinstance(t, sos_targets):
                raise RuntimeError(
                    f"Nested sos_targets {t} were introduced by {args}")
            if not isinstance(t, BaseTarget):
                raise RuntimeError(f"Unrecognized target {t}")
        if _verify_existence:
            for target in self._targets:
                if not target.target_exists('any'):
                    raise UnknownTarget(target)
        if group_by:
            self._group(group_by)
        if paired_with:
            self._handle_paired_with(paired_with)
        if pattern:
            self._handle_extract_pattern(pattern)
        if group_with:
            self._handle_group_with(group_with)
        if for_each:
            self._handle_for_each(for_each)

    def set_traced(self):
        [x.set_traced() for x in self._targets]
        self.traced = True
        return self

    def is_external(self):
        if not self.valid():
            return False
        return all(x.is_external()
                   for x in self._targets
                   if isinstance(x, file_target))

    def unspecified(self):
        return not self._targets and self._undetermined is True

    def undetermined(self):
        return not self._targets and isinstance(self._undetermined, str)

    def touch(self):
        for x in self._targets:
            if isinstance(x, file_target):
                x.touch()

    def valid(self):
        return self._targets or self._undetermined is False

    def __append__(self,
                   arg,
                   source='',
                   default_source='',
                   verify_existence=False):
        src = source if source else default_source
        if isinstance(arg, paths):
            self._targets.extend([file_target(x) for x in arg._paths])
            self._labels.extend([src] * len(arg._paths))
            for g in self._groups:
                g.add_last(len(arg._paths), parent=self)
        elif isinstance(arg, path):
            self._targets.append(file_target(arg))
            self._labels.append(src)
            for g in self._groups:
                g.add_last(1, parent=self)
        elif isinstance(arg, str):
            if self.wildcard.search(arg):
                matched = sorted(glob.glob(os.path.expanduser(arg)))
                if matched:
                    self._targets.extend([file_target(x) for x in matched])
                    self._labels.extend([src] * len(matched))
                    for g in self._groups:
                        g.add_last(len(matched), parent=self)
                elif verify_existence:
                    raise UnknownTarget(arg)
                else:
                    env.logger.debug(f'Pattern {arg} does not match any file')
            else:
                self._targets.append(file_target(arg))
                self._labels.append(src)
                for g in self._groups:
                    g.add_last(1, parent=self)
        elif isinstance(arg, dict):
            for k, v in arg.items():
                if not isinstance(k, str):
                    raise ValueError(
                        f'Failed to create a sos_targets object with dictionary {arg}: source of sos_targets can only be a string: {k} of type {k.__class__.__name__} specified'
                    )
                self.__append__(v, source=k, verify_existence=verify_existence)
        elif isinstance(arg, sos_targets):
            # group handling if tricker here because arg can have
            # its own groups.
            if source:
                self.extend(arg, source=source)
            else:
                # use the source from sos_target, not from default
                self.extend(arg)
        elif isinstance(arg, BaseTarget):
            self._targets.append(arg)
            self._labels.append(src)
            for g in self._groups:
                g.add_last(1, parent=self)
        elif isinstance(arg, Iterable):
            # in case arg is a Generator, check its type will exhaust it
            for t in list(arg):
                self.__append__(t, source=src)
        elif arg is not None:
            raise RuntimeError(
                f'Unrecognized targets {arg} of type {arg.__class__.__name__}')

    def set_labels(self, source):
        if isinstance(source, str):
            self._labels = [source] * len(self._targets)
        elif len(source) == len(self._targets):
            self._labels = source
        else:
            raise ValueError(
                f'Invalid source {source} for sos_target with {len(self)} targets.'
            )

    labels = property(lambda self: self._labels, set_labels)

    targets = property(lambda self: self._targets)

    groups = property(
        lambda self: [x.idx_to_targets(self) for x in self._groups])

    def _get_group(self, index):
        return self._groups[index].idx_to_targets(self)

    #def targets(self):
    #    return [x.target_name() if isinstance(x, file_target) else x for x in self._targets]

    def later_than(self, other):
        '''if the current target is later than the other targets'''
        # if no input, output is naturally later
        file_rhs = [x for x in other.targets if isinstance(x, file_target)]
        if not file_rhs:
            return True
        # if no output, but input, we cannot skip
        file_lhs = [x for x in self._targets if isinstance(x, file_target)]
        if not file_lhs:
            return False
        # now we have both.
        return min(x.stat().st_mtime for x in file_lhs) > max(
            x.stat().st_mtime for x in file_rhs)

    def extend(self, another, source='', keep_groups=False):
        if isinstance(another, sos_targets):
            arg = another
        else:
            arg = sos_targets(another, _source=source)
        if arg.valid() and not self.valid():
            self._undetermined = False
        #
        n_old = len(self._targets)
        n_added = len(arg._targets)

        self._targets.extend(arg._targets)
        # if source is specified, override the default
        if source:
            self._labels.extend([source] * len(arg._targets))
        elif hasattr(arg, '_labels'):
            self._labels.extend(arg._labels)
        else:
            self._labels.extend([''] * len(arg._targets))
        # merge dictionaries
        self._dict.update(arg._dict)
        #
        if keep_groups:
            return self
        # it is possible to merge groups from multiple...
        if arg._groups:
            # if source is specified, it will override labels of all groups
            if not self._groups:
                self._groups = [
                    _sos_group(range(n_old), parent=self) for g in arg._groups
                ]
            if len(self._groups) == 1 and len(arg._groups) > 1:
                # 1 vs more, we duplicate itself
                self._groups = [
                    _sos_group(
                        self._groups[0]._indexes,
                        self._groups[0]._labels).set(**self._groups[0]._dict)
                    for ag in arg._groups
                ]
                for g, ag in zip(self._groups, arg._groups):
                    g.extend(ag, start=n_old, parent=self)
            elif len(self._groups) > 1 and len(arg._groups) == 1:
                for g in self._groups:
                    g.extend(arg._groups[0], start=n_old, parent=self)
            elif len(self._groups) == len(arg._groups):
                for g, ag in zip(self._groups, arg._groups):
                    g.extend(ag, start=n_old, parent=self)
            else:
                raise ValueError(
                    f'Cannot merge a sos_targets objects with {len(self._groups)} groups with another sos_targets object with {len(ag)} groups.'
                )
        elif self._groups:
            # if the RHS has no _group but myself has groups...
            # source will be figured out during extend
            ag = _sos_group(range(n_added), labels=[''] * n_added)
            for g in self._groups:
                g.extend(ag, start=n_old, parent=self)
        return self

    def zap(self):
        for target in self._targets:
            if isinstance(target, file_target):
                target.zap()
            else:
                env.logger.debug(f'Ignore non-file target {target}')

    def __getstate__(self):
        return (self._targets, self._labels, self._undetermined, self._groups,
                self._dict)

    def __setstate__(self, state) -> None:
        if isinstance(state, tuple):
            if len(state) == 2:
                self._targets = state[0]
                self._labels = [''] * len(self._targets)
                self._undetermined = state[1]
                self._groups = []
                self._dict = {}
            elif len(state) == 3:
                self._targets = state[0]
                self._labels = state[1]
                self._undetermined = state[2]
                self._groups = []
                self._dict = {}
            elif len(state) == 4:
                self._targets = state[0]
                self._labels = state[1]
                self._undetermined = state[2]
                self._groups = state[3]
                self._dict = {}
            elif len(state) == 5:
                self._targets = state[0]
                self._labels = state[1]
                self._undetermined = state[2]
                self._groups = state[3]
                self._dict = state[4]
        else:
            # older version of sig file might only saved targets
            self._targets = state
            self._labels = [''] * len(self._targets)
            self._undetermined = False
            self._groups = []
            self._dict = {}

    def __len__(self):
        return len(self._targets)

    def select(self, i):
        # similar to [] but always returns a sos_targets object with appropriate source
        if isinstance(i, str):
            ret = sos_targets()
            ret._undetermined = self._undetermined
            ret._targets = [
                x for x, y in zip(self._targets, self._labels) if y == i
            ]
            index_map = {
                o_idx: n_idx for n_idx, o_idx in zip(
                    range(len(ret._targets)),
                    [x for x, y in enumerate(self._labels) if y == i])
            }
            ret._labels = [i] * len(ret._targets)
            ret._groups = []
            for grp in self._groups:
                ret._groups.append(
                    _sos_group([
                        index_map[x]
                        for x, y in zip(grp._indexes, grp._labels)
                        if y == i
                    ],
                               labels=i).set(**grp._dict))
            return ret
        elif isinstance(i, (tuple, list)):
            ret = sos_targets()
            ret._undetermined = self._undetermined
            ret._targets = [self._targets[x] for x in i]
            ret._labels = [self._labels[x] for x in i]
            ret._groups = []
            return ret
        elif callable(i):
            kept = [idx for idx, x in enumerate(self._targets) if i(x)]
            if len(kept) == len(self._targets):
                return self
            ret = sos_targets()
            ret._undetermined = self._undetermined
            ret._targets = [self._targets[x] for x in kept]
            ret._labels = [self._labels[x] for x in kept]
            ret._groups = []
            if not self._groups:
                return ret
            index_map = {
                o_idx: n_idx
                for n_idx, o_idx in zip(range(len(ret._targets)), kept)
            }
            kept = set(kept)
            for grp in self._groups:
                ret._groups.append(
                    _sos_group(
                        [index_map[x] for x in grp._indexes if x in kept], [
                            y for x, y in zip(grp._indexes, grp._labels)
                            if x in kept
                        ]).set(**grp._dict))
            return ret
        else:
            ret = sos_targets()
            ret._undetermined = self._undetermined
            ret._targets = [self._targets[i]] if isinstance(
                i, int) else self._targets[i]
            ret._labels = [self._labels[i]] if isinstance(
                i, int) else self._labels[i]
            ret._groups = []
            return ret

    def __getitem__(self, i):
        if isinstance(i, str):
            ret = sos_targets()
            ret._undetermined = self._undetermined
            ret._targets = [
                x for x, y in zip(self._targets, self._labels) if y == i
            ]
            index_map = {
                o_idx: n_idx for n_idx, o_idx in zip(
                    range(len(ret._targets)),
                    [x for x, y in enumerate(self._labels) if y == i])
            }
            ret._labels = [i] * len(ret._targets)
            ret._groups = []
            for grp in self._groups:
                ret._groups.append(
                    _sos_group([
                        index_map[x]
                        for x, y in zip(grp._indexes, grp._labels)
                        if y == i
                    ],
                               labels=i).set(**grp._dict))
            return ret
        else:
            return self._targets[i]

    def target_signature(self):
        return tuple((x.target_signature(), y)
                     for x, y in zip(self._targets, self._labels))

    def validate(self, sig):
        return isinstance(sig, tuple) and len(sig) == len(
            self._targets) and all(
                x.validate(sig[0]) and src == sig[1]
                for x, src, sig in zip(self._targets, self._labels, sig))

    def target_exists(self, mode='any'):
        return all(x.target_exists(mode) for x in self._targets)

    def __getattr__(self, name):
        try:
            return self._dict[name]
        except Exception:
            if len(self._targets) == 1:
                try:
                    return getattr(self._targets[0], name)
                except Exception:
                    raise AttributeError(
                        f'{self.__class__.__name__} object or its first child has no attribute {name}'
                    )
            else:
                raise AttributeError(
                    f'{self.__class__.__name__} object has no attribute {name}')

    def target_name(self):
        return f"sos_targets([{','.join(x.target_name() for x in self._targets)}],_labels=[{','.join(self._labels)}])"

    def _dedup(self):
        kept = []
        items = set()
        for i, t in enumerate(self._targets):
            if t not in items:
                kept.append(i)
                items.add(t)
        return self.remove_targets(self, kept=kept)

    def paired_with(self, name, properties):
        # can pair with sos_targets
        if not isinstance(properties,
                          sos_targets) and not is_basic_type(properties):
            env.logger.warning(
                f'Failed to paired_with with value "{properties}" as it contains unsupported data type'
            )
            return self
        if isinstance(properties, (bool, int, float, str, bytes)):
            for target in self._targets:
                target.set(name, properties)
        elif isinstance(properties, (list, tuple, sos_targets)):
            if len(properties) != len(self._targets):
                raise ValueError(
                    f'Length of provided attributes ({len(properties)}) does not match length of sos_targets ({len(self._targets)})'
                )
            for target, property in zip(self._targets, properties):
                target.set(name, property)
        else:
            raise ValueError(
                'Unacceptable properties {properties} for function paired_with')
        return self

    def remove_targets(self, type, kept=None):
        '''Remove targets of certain type'''
        if kept is None:
            kept = [
                i for i, x in enumerate(self._targets)
                if not isinstance(x, type)
            ]
        if len(kept) == len(self._targets):
            return self
        self._targets = [self._targets[x] for x in kept]
        self._labels = [self._labels[x] for x in kept]
        if not self._groups:
            return self
        index_map = {
            o_idx: n_idx
            for n_idx, o_idx in zip(range(len(self._targets)), kept)
        }
        kept = set(kept)
        for idx, grp in enumerate(self._groups):
            self._groups[idx] = _sos_group(
                [index_map[x] for x in grp._indexes if x in kept],
                [y for x, y in zip(grp._indexes, grp._labels) if x in kept
                ]).set(**grp._dict)
        return self

    def resolve_remote(self):
        '''If target is of remote type, resolve it'''
        for idx, target in enumerate(self._targets):
            if isinstance(target, remote):
                resolved = target.resolve()
                if isinstance(resolved, str):
                    resolved = interpolate(resolved, env.sos_dict.dict())
                self._targets[idx] = file_target(resolved).set(**target._dict)
        return self

    def group_with(self, name, properties):
        if not self._groups:
            self._group(by='all')
        if not is_basic_type(properties):
            env.logger.warning(
                f'Failed to set {properties} as it is or contains unsupported data type'
            )
            return self
        if isinstance(properties, (bool, int, float, str, bytes)):
            for group in self._groups:
                group.set(name, properties)
        elif isinstance(properties, (list, tuple)):
            if len(properties) != len(self._groups):
                raise ValueError(
                    f'Length of provided properties ({len(properties)}) does not match number of groups ({len(self._groups)})'
                )
            for group, property in zip(self._groups, properties):
                group.set(name, property)
        else:
            raise ValueError(
                'Unacceptable properties {properties} of type {properties.__class__.__name__} for function group_with'
            )
        return self

    def get(self, name, default=None):
        if name in self._dict:
            return self._dict[name]
        elif len(self._targets) == 1:
            return self._targets[0].get(name, default)
        else:
            return default

    def _add_groups(self, grps):
        self._groups = []

        for grp in grps:
            if not isinstance(grp, sos_targets):
                raise RuntimeError(
                    f'_output should be of type sos_targets: {grp} of type {grp.__class__.__name__} returned.'
                )
            start_idx = len(self._targets)
            grp_size = len(grp)
            self._targets.extend(grp._targets)
            self._labels.extend(grp._labels)
            self._groups.append(
                _sos_group(
                    range(start_idx, start_idx + grp_size),
                    labels=grp._labels).set(**grp._dict))
        # in theory the groups should not overlap but in rare cases when
        # output is for example dynamic, they could overlap.
        return self._dedup()

    def _remove_empty_groups(self):
        self._groups = [x for x in self._groups if len(x._indexes) > 0]
        return self

    def _duplicate_groups(self, n):
        n_grps = len(self._groups)
        for _ in range(n - 1):
            for grp in self._groups[:n_grps]:
                self._groups.append(
                    _sos_group(grp._indexes, grp._labels).set(**grp._dict))
        return self

    def _num_groups(self):
        return len(self._groups)

    def _group(self, by):
        if by is None:
            return self

        if self._groups:
            self._groups = []

        if by == 'single':
            self._groups = [
                _sos_group([x], parent=self) for x in range(len(self))
            ]
        elif by == 'all':
            # default option
            self._groups = [_sos_group(range(len(self)), self._labels)]
        elif isinstance(by, str) and (by.startswith('pairsource') or by.startswith('pairlabel')):
            labels = list(dict.fromkeys(self.labels))
            if len(labels) == 1:
                raise ValueError(
                    f'Cannot pairlabel input with a single label.')
            if by == 'pairsource' or by == 'pairlabel':
                grp_size = 1
            else:
                try:
                    if by.startswith('pairsource'):
                        grp_size = int(by[10:])
                    else:
                        grp_size = int(by[9:])
                except Exception:
                    raise ValueError(f'Invalid pairsource option {by}')
            src_sizes = {s: self.labels.count(s) for s in labels}
            if max(src_sizes.values()) % grp_size != 0:
                raise ValueError(
                    f'Cannot use group size {grp_size} (option {by}) for source of size {src_sizes}'
                )
            n_groups = max(src_sizes.values()) // grp_size
            indexes = [[] for x in range(n_groups)]
            for s in labels:
                lookup = [
                    idx for idx, src in enumerate(self.labels) if src == s
                ]
                if src_sizes[s] > n_groups and src_sizes[s] % n_groups == 0:
                    gs = src_sizes[s] // n_groups
                    for i in range(n_groups):
                        # (0, 1, 2), (3, 4, 5), (6, 7, 8) ...
                        indexes[i].extend(lookup[i * gs:(i + 1) * gs])
                elif n_groups >= src_sizes[s] and n_groups % src_sizes[s] == 0:
                    for i in range(n_groups):
                        # (0 ), (0, ), (1, ), (1, ) ...
                        indexes[i].append(lookup[i //
                                                 (n_groups // src_sizes[s])])
                else:
                    raise ValueError(
                        f'Cannot use group size {grp_size} (by="{by}") for source of size {src_sizes}'
                    )
            self._groups = [
                _sos_group(indexes[x], parent=self) for x in range(n_groups)
            ]
        elif isinstance(by, str) and by.startswith('pairs'):
            if len(self) % 2 != 0:
                raise ValueError(
                    f'Paired by has to have even number of input files: {len(self)} provided'
                )
            if by == 'pairs':
                grp_size = 1
            else:
                try:
                    grp_size = int(by[5:])
                except Exception:
                    raise ValueError(f'Invalid pairs option {by}')
            if grp_size == 1:
                self._groups = [
                    _sos_group(x, parent=self) for x in zip(
                        range(0,
                              len(self) // 2), range(len(self) // 2, len(self)))
                ]
            else:
                if len(self) % grp_size != 0:
                    raise ValueError(
                        f'Paired by with group size {grp_size} is not possible with input of size {len(self)}'
                    )
                self._groups = [
                    _sos_group(
                        list(range(x[0], x[0] + grp_size)) +
                        list(range(x[1], x[1] + grp_size)),
                        parent=self) for x in zip(
                            range(0,
                                  len(self) // 2, grp_size),
                            range(len(self) // 2, len(self), grp_size))
                ]
        elif isinstance(by, str) and by.startswith('pairwise'):
            if by == 'pairwise':
                grp_size = 1
            else:
                try:
                    grp_size = int(by[8:])
                except Exception:
                    raise ValueError(f'Invalid pairs option {by}')
            if grp_size == 1:
                f1, f2 = tee(range(len(self)))
                next(f2, None)
                self._groups = [_sos_group(x, parent=self) for x in zip(f1, f2)]
            else:
                if len(self) % grp_size != 0:
                    raise ValueError(
                        f'Paired by with group size {grp_size} is not possible with input of size {len(self)}'
                    )
                f1, f2 = tee(range(len(self) // grp_size))
                next(f2, None)
                self._groups = [
                    _sos_group(
                        list(range(x[0] * grp_size, (x[0] + 1) * grp_size)) +
                        list(range(x[1] * grp_size, (x[1] + 1) * grp_size)),
                        parent=self) for x in zip(f1, f2)
                ]
        elif isinstance(by, str) and by.startswith('combinations'):
            if by == 'combinations':
                grp_size = 2
            else:
                try:
                    grp_size = int(by[12:])
                except Exception:
                    raise ValueError(f'Invalid pairs option {by}')
            self._groups = [
                _sos_group(x, parent=self)
                for x in combinations(range(len(self)), grp_size)
            ]
        elif by == 'source' or by == 'label':
            labels = list(dict.fromkeys(self.labels))
            self._groups = [
                _sos_group([i for i, x in enumerate(self._labels) if x == src],
                           parent=self) for src in labels
            ]
        elif isinstance(by, int) or (isinstance(by, str) and by.isdigit()):
            by = int(by)
            if len(self) % by != 0 and len(self) > by:
                env.logger.warning(
                    f'Number of samples ({len(self)}) is not a multiple of by ({by}). The last group would have less files than the other groups.'
                )
            if by < 1:
                raise ValueError(
                    'Value of paramter by should be a positive number.')
            self._groups = [
                _sos_group(range(i, min(i + by, len(self))), parent=self)
                for i in range(0, len(self), by)
            ]
        elif callable(by):
            try:
                self._groups = []
                idx = by(self)
                try:
                    idx = list(idx)
                except Exception:
                    raise ValueError(
                        f'Customized grouping method should return a list. {idx} of type {idx.__class__.__name__} is returned.'
                    )
                for grp in by(self):
                    if isinstance(grp, Sequence) and all(
                            isinstance(x, int) for x in grp):
                        if any(x < 0 or x >= len(self._targets) for x in grp):
                            raise ValueError(
                                f'Index out of range (< {len(self._targets)}): {grp}'
                            )
                        self._groups.append(_sos_group(grp, parent=self))
                    else:
                        index = []
                        for x in sos_targets(grp):
                            try:
                                index.append(self._targets.index(x))
                            except Exception:
                                raise ValueError(
                                    f'Returned target is not one of the targets. {x}'
                                )
                        self._groups.append(_sos_group(index, parent=self))
            except Exception as e:
                raise ValueError(
                    f'Failed to apply customized grouping method: {e}')
        else:
            raise ValueError(f'Unsupported by option ``{by}``!')
        return self

    def _handle_paired_with(self, paired_with):
        '''Handle input option paired_with'''
        if paired_with is None or not paired_with:
            var_name = []
            var_value = []
        elif isinstance(paired_with, str):
            var_name = ['_' + paired_with]
            if paired_with not in env.sos_dict:
                raise ValueError(f'Variable {paired_with} does not exist.')
            var_value = [env.sos_dict[paired_with]]
        elif isinstance(paired_with, dict):
            var_name = []
            var_value = []
            for k, v in paired_with.items():
                var_name.append(k)
                var_value.append(v)
        elif isinstance(paired_with, Iterable):
            try:
                var_name = ['_' + x for x in paired_with]
            except Exception:
                raise ValueError(
                    f'Invalud value for option paired_with {paired_with}')
            var_value = []
            for vn in var_name:
                if vn[1:] not in env.sos_dict:
                    raise ValueError(f'Variable {vn[1:]} does not exist.')
                var_value.append(env.sos_dict[vn[1:]])
        else:
            raise ValueError(
                f'Unacceptable value for parameter paired_with: {paired_with}')
        #
        for vn, vv in zip(var_name, var_value):
            # set paired with values to step_input
            self.paired_with(vn, vv)

    def _handle_group_with(self, group_with):
        '''Handle input option group_with'''
        if group_with is None or not group_with:
            var_name = []
            var_value = []
        elif isinstance(group_with, str):
            var_name = ['_' + group_with]
            if group_with not in env.sos_dict:
                raise ValueError(f'Variable {group_with} does not exist.')
            var_value = [env.sos_dict[group_with]]
        elif isinstance(group_with, dict):
            var_name = []
            var_value = []
            for k, v in group_with.items():
                var_name.append(k)
                var_value.append(v)
        elif isinstance(group_with, Iterable):
            try:
                var_name = ['_' + x for x in group_with]
            except Exception:
                raise ValueError(
                    f'Invalud value for option group_with {group_with}')
            var_value = []
            for vn in var_name:
                if vn[1:] not in env.sos_dict:
                    raise ValueError(f'Variable {vn[1:]} does not exist.')
                var_value.append(env.sos_dict[vn[1:]])
        else:
            raise ValueError(
                f'Unacceptable value for parameter group_with: {group_with}')
        #
        for vn, vv in zip(var_name, var_value):
            self.group_with(vn, vv)

    def _handle_extract_pattern(self, pattern):
        '''Handle input option pattern'''
        if pattern is None or not pattern:
            patterns = []
        elif isinstance(pattern, str):
            patterns = [pattern]
        elif isinstance(pattern, Iterable):
            patterns = pattern
        else:
            raise ValueError(
                f'Unacceptable value for parameter pattern: {pattern}')
        #
        for pattern in patterns:
            res = extract_pattern(pattern, self._targets)
            self.set(**res)
            # also make k, v pair with _input
            self._handle_paired_with({'_' + x: y for x, y in res.items()})

    def _handle_for_each(self, for_each):
        if for_each is None or not for_each:
            for_each = []
        elif isinstance(for_each, (str, dict)):
            for_each = [for_each]
        elif isinstance(for_each, Sequence):
            for_each = for_each
        else:
            raise ValueError(
                f'Unacceptable value for parameter for_each: {for_each}')
        #
        for fe_all in for_each:
            if isinstance(fe_all, dict):
                # in the format of {'name': value}
                fe_iter_names = []
                fe_values = []
                for k, v in fe_all.items():
                    if ',' in k:
                        names = [x.strip() for x in k.split(',')]
                        if isinstance(v, Iterable):
                            v = list(v)
                        if any(len(_v) != len(names) for _v in v):
                            raise ValueError(
                                f'Unable to unpack object {short_repr(v)} for variables {k} (of length {len(names)})'
                            )
                        fe_iter_names.extend(names)
                        fe_values.extend(list(zip(*v)))
                    else:
                        fe_iter_names.append(k)
                        fe_values.append(v)
            else:
                if ',' in fe_all:
                    fe_var_names = [x.strip() for x in fe_all.split(',')]
                    fe_iter_names = ['_' + x for x in fe_var_names]
                else:
                    fe_var_names = [fe_all]
                    fe_iter_names = ['_' + fe_all]
                # check iterator variable name
                for name in fe_iter_names:
                    if '.' in name:
                        raise ValueError(f'Invalid iterator variable {name}')
                # check variables
                fe_values = []
                for name in fe_var_names:
                    if name.split('.')[0] not in env.sos_dict:
                        raise ValueError(f'Variable {name} does not exist.')
                    if '.' in name:
                        fe_values.append(
                            getattr(env.sos_dict[name.split('.')[0]],
                                    name.split('.', 1)[-1]))
                    else:
                        fe_values.append(env.sos_dict[name])

            # get loop size
            loop_size = None
            for name, values in zip(fe_iter_names, fe_values):
                if not isinstance(values, Sequence):
                    try:
                        import pandas as pd
                        if not isinstance(values,
                                          (pd.DataFrame, pd.Series, pd.Index)):
                            raise ValueError(
                                f'Unacceptable for_each data type {values.__class__.__name__}'
                            )
                    except Exception as e:
                        raise ValueError(
                            f'Cannot iterate through variable {name}: {e}')
                if loop_size is None:
                    loop_size = len(values)
                elif loop_size != len(values):
                    raise ValueError(
                        f'Length of variable {name} (length {len(values)}) should match the length of other variables (length {loop_size}).'
                    )

            n_grps = self._num_groups()
            if n_grps == 0:
                self._group('all')
                n_grps = 1
            self._duplicate_groups(loop_size)
            #
            for vidx in range(loop_size):
                for idx in range(n_grps):
                    for var_name, values in zip(fe_iter_names, fe_values):
                        if isinstance(values, Sequence):
                            self._groups[n_grps * vidx + idx].set(
                                var_name, values[vidx])
                        elif isinstance(values, pd.DataFrame):
                            self._groups[n_grps * vidx + idx].set(
                                var_name, values.iloc[vidx])
                        elif isinstance(values, pd.Series):
                            self._groups[n_grps * vidx + idx].set(
                                var_name, values.iloc[vidx])
                        elif isinstance(values, pd.Index):
                            self._groups[n_grps * vidx + idx].set(
                                var_name, values[vidx])
                        else:
                            raise ValueError(
                                f'Failed to iterate through for_each variable {short_repr(values)}'
                            )

    def __hash__(self):
        return hash(repr(self))

    def __eq__(self, other):
        try:
            # allow compare to any object as long as it can be converted to sos_targets
            return self._targets == (
                other._targets if isinstance(other, sos_targets) else
                sos_targets(other)._targets)
        except Exception:
            return False

    def __add__(self, part):
        if len(self._targets) == 1:
            return self._targets[0].__add__(part)
        elif len(self._targets) == 0:
            raise ValueError(f"Cannot add {part} to empty target list")
        else:
            raise ValueError(
                f'Cannot add {part} to group of {len(self)} targets {self!r}')

    def __fspath__(self):
        if len(self._targets) == 1:
            return self._targets[0].__fspath__()
        elif len(self._targets) == 0:
            raise ValueError(
                f"Cannot treat an empty sos_targets as single target")
        else:
            raise ValueError(
                f'Cannot treat an sos_targets object {self} with more than one targets as a single target'
            )

    def __repr__(self):
        return ('[' + ', '.join(repr(x) for x in self._targets) +
                ']') if self.valid() else (
                    'Unspecified' if self.unspecified() else self._undetermined)

    def __short_repr__(self):
        grp_info = '' if self._num_groups(
        ) <= 1 else f' in {self._num_groups()} groups'
        if self.valid():
            if len(self._targets) <= 2:
                return ' '.join([x.target_name() for x in self._targets
                                ]) + grp_info
            else:
                return ' '.join([
                    x.target_name() for x in self._targets[:2]
                ]) + f'... ({len(self._targets)} items{grp_info})'
        else:
            return 'Unspecified' if self.unspecified() else self._undetermined

    def __stable_repr__(self):
        return repr(self)

    def __str__(self):
        return self.__format__('') if self.valid() else (
            'Unspecified' if self.unspecified() else self._undetermined)

    def __format__(self, format_spec):
        if not self.valid():
            return 'Unspecified' if self.unspecified() else self._undetermined
        if ',' in format_spec:
            fmt_spec = format_spec.replace(',', '')
            return ','.join(x.__format__(fmt_spec) for x in self._targets)
        else:
            return ' '.join(x.__format__(format_spec) for x in self._targets)

    def __deepcopy__(self, memo):
        ret = sos_targets()
        ret._targets = copy.deepcopy(self._targets)
        ret._groups = copy.deepcopy(self._groups)
        ret._labels = copy.deepcopy(self._labels)
        ret._dict = copy.deepcopy(self._dict)
        ret._undetermined = self._undetermined
        return ret

    def contains(self, target):
        if isinstance(target, str):
            return file_target(target) in self._targets
        else:
            return target in self._targets


class InMemorySignature:

    def __init__(self,
                 input_files: sos_targets,
                 output_files: sos_targets,
                 dependent_files: sos_targets,
                 signature_vars: set = set(),
                 sdict: dict = {},
                 shared_vars: list = []):
        '''Runtime information for specified output files
        '''
        self.content = None
        if not sdict:
            sdict = env.sos_dict
        if not input_files.valid():
            raise RuntimeError(
                'Input files of step signature cannot be undetermined.')
        if not dependent_files.valid():
            raise RuntimeError(
                'Dependent files of step signature cannot be undetermined.')

        self.input_files = input_files.remove_targets(type=sos_step)
        self.dependent_files = dependent_files.remove_targets(type=sos_step)
        self.output_files = output_files.remove_targets(type=sos_step)
        self.signature_vars = signature_vars
        self.shared_vars = shared_vars
        # signatures that exist before execution and might change during execution
        self.init_signature = {
            x: deepcopy(sdict[x]) for x in sorted(signature_vars) if
            x in sdict and not callable(sdict[x]) and pickleable(sdict[x], x)
        }

    def write(self):
        if self.content is not None:
            return self.content
        if self.output_files.undetermined():
            self.output_files = env.sos_dict['_output']
            env.log_to_file(
                'TARGET',
                f'Set undetermined output files to {env.sos_dict["_output"]}')
        input_sig = {}
        for f in self.input_files:
            try:
                input_sig[str(f)] = f.target_signature()
            except Exception:
                env.logger.debug(
                    f'Failed to create signature: input target {f} does not exist'
                )
                return False
        output_sig = {}
        for f in self.output_files:
            try:
                output_sig[str(f)] = f.target_signature()
            except Exception:
                env.logger.debug(
                    f'Failed to create signature: output target {f} does not exist'
                )
                return False
        dependent_sig = {}
        for f in self.dependent_files:
            try:
                dependent_sig[str(f)] = f.target_signature()
            except Exception:
                env.logger.debug(
                    f'Failed to create signature: dependent target {f} does not exist'
                )
                return False
        init_context_sig = {
            var: objectMD5(self.init_signature[var])
            for var in self.init_signature
            if pickleable(self.init_signature[var], var)
        }
        if self.shared_vars:
            end_context = {
                var: env.sos_dict[var]
                for var in self.shared_vars
                if var in env.sos_dict and pickleable(env.sos_dict[var], var)
            }
        else:
            end_context = {}

        self.content = {
            'input': input_sig,
            'input_obj': self.input_files,
            'output': output_sig,
            'output_obj': self.output_files,
            'depends': dependent_sig,
            'depends_obj': self.dependent_files,
            'init_context_sig': init_context_sig,
            'end_context': end_context
        }
        return self.content

    def validate(self, signature):
        '''Check if ofiles and ifiles match signatures recorded in md5file'''
        if not signature:
            return 'Empty signature'
        sig_files = self.input_files._targets + self.output_files._targets + \
            self.dependent_files._targets
        for x in sig_files:
            if not x.target_exists('any'):
                return f'Missing target {x}'
        #

        files_checked = {x.target_name(): False for x in sig_files}
        res = {'input': [], 'output': [], 'depends': [], 'vars': {}}
        cur_type = 'input'
        # old signature
        if 'init_context' in signature:
            for key, value in signature['init_context'].items():
                if key not in env.sos_dict:
                    return f'Variable {key} not in running environment'
                try:
                    if env.sos_dict[key] != value:
                        return f'Context variable {key} value mismatch: {short_repr(value)} saved, {short_repr(env.sos_dict[key])} current'
                except Exception as e:
                    env.logger.debug(
                        f"Variable {key} of type {type(value).__name__} cannot be compared: {e}"
                    )
        elif 'init_context_sig' in signature:
            for key, value in signature['init_context_sig'].items():
                if key not in env.sos_dict:
                    return f'Variable {key} not in running environment'
                try:
                    if objectMD5(env.sos_dict[key]) != value:
                        return f'ID of context variable {key} ({objectMD5(env.sos_dict[key])}) mismatch: {short_repr(env.sos_dict[key])} does not match id {value}'
                except Exception as e:
                    env.logger.debug(f"Variable {key} cannot be compared: {e}")

        res['vars'].update(signature['end_context'])
        #
        for cur_type in ['input', 'output', 'depends']:
            for f, m in signature[cur_type].items():
                try:
                    if '(' in f and ')' in f:
                        # this part is hard, because this can be a customized target.
                        target_type = f.split('(')[0]
                        target_class = None
                        if target_type in globals():
                            target_class = eval(target_type)
                        else:
                            # check registry
                            for entrypoint in pkg_resources.iter_entry_points(
                                    group='sos_targets'):
                                if entrypoint.name.strip() == target_type:
                                    target_class = entrypoint.load()
                                    break
                        if target_class is None:
                            raise ValueError(
                                f'Failed to identify target class {target_type}'
                            )
                        # parameter of class?
                        freal = eval(f, {target_type: target_class})
                    else:
                        freal = file_target(f)
                    if not freal.validate(m):
                        return f'Target {f} does not exist or does not match saved signature {m}'
                    res[cur_type].append(freal.target_name(
                    ) if isinstance(freal, file_target) else freal)
                    files_checked[freal.target_name()] = True
                except Exception as e:
                    env.logger.debug(f'Wrong md5 in signature: {e}')
        #
        if not all(files_checked.values()):
            return f'No MD5 signature for {", ".join(x for x,y in files_checked.items() if not y)}'
        if 'input_obj' in signature:
            # for new style signature, the entire objects are kept
            res['input'] = signature['input_obj']
            res['depends'] = signature['depends_obj']
            res['output'] = signature['output_obj']
        return res


class RuntimeInfo(InMemorySignature):
    '''Record run time information related to a number of output files. Right now only the
    .exe_info files are used.
    '''

    def __init__(self,
                 step_md5: str,
                 input_files: sos_targets,
                 output_files: sos_targets,
                 dependent_files: sos_targets,
                 signature_vars: set = set(),
                 sdict: dict = {},
                 shared_vars: list = []):
        '''Runtime information for specified output files
        '''
        if 'sos_run' in signature_vars:
            # if a step has nested workflow, we cannot save signature
            # because we do not know the exact content of the nested workflow.
            self.sig_id = ''
            return

        if not sdict:
            sdict = env.sos_dict
        self.step_md5 = step_md5
        super(RuntimeInfo, self).__init__(
            input_files,
            output_files,
            dependent_files,
            signature_vars,
            shared_vars=shared_vars)

        self.sig_id = textMD5(
            f'{self.step_md5} {self.input_files} {self.output_files} {self.dependent_files} {stable_repr(self.init_signature)}{sdict["_index"] if self.output_files.undetermined() else ""}'
        )

    def __getstate__(self):
        if not self.sig_id:
            return {}
        return {
            'step_md5': self.step_md5,
            'input_files': self.input_files,
            'output_files': self.output_files,
            'dependent_files': self.dependent_files,
            'signature_vars': self.signature_vars,
            'init_signature': self.init_signature,
            'sig_id': self.sig_id
        }

    def __setstate__(self, sdict: Dict[str, Any]):
        if not sdict:
            self.sig_id = ''
            return
        self.step_md5 = sdict['step_md5']
        self.input_files = sdict['input_files']
        self.output_files = sdict['output_files']
        self.dependent_files = sdict['dependent_files']
        self.signature_vars = sdict['signature_vars']
        self.init_signature = sdict['init_signature']
        self.sig_id = sdict['sig_id']

    def lock(self):
        if not self.sig_id:
            return
        # we will need to lock on a file that we do not really write to
        # otherwise the lock will be broken when we write to it.
        self._lock = fasteners.InterProcessLock(
            os.path.join(env.temp_dir, self.sig_id + '.lock'))
        if not self._lock.acquire(blocking=False):
            self._lock = None
            raise UnavailableLock((self.input_files, self.output_files,
                                   os.path.join(env.temp_dir,
                                                self.sig_id + '.lock')))
        else:
            env.log_to_file(
                'TARGET',
                f'Lock acquired for output files {short_repr(self.output_files)}'
            )

    def release(self, quiet=False):
        if not self.sig_id:
            return
        if not hasattr(self, '_lock') or self._lock is None:
            env.logger.warning(
                f'Releasing an non-existent or released lock for {self.sig_id}.'
            )
            return
        if self._lock:
            try:
                self._lock.release()
                env.log_to_file(
                    'TARGET',
                    f'Lock released for output files {short_repr(self.output_files)}'
                )
            except Exception as e:
                if not quiet:
                    env.logger.warning(
                        f'Unable to release lock for output files {self.output_files}: {e}'
                    )
            finally:
                self._lock = None

    def set_output(self, files: sos_targets):
        if not self.sig_id:
            return
        # add signature file if input and output files are dynamic
        if 'TARGET' in env.config['SOS_DEBUG'] or 'ALL' in env.config[
                'SOS_DEBUG']:
            env.log_to_file('TARGET', f'Set output of signature to {files}')
        self.output_files = files

    def write(self):
        '''Write signature file with signature of script, input, output and dependent files.
        Because local input and output files can only be determined after the execution
        of workflow. They are not part of the construction.
        '''
        if not self.sig_id:
            return False
        if not self.output_files.valid():
            raise ValueError(
                f'Cannot write signature with undetermined output {self.output_files}'
            )
        else:
            if 'TARGET' in env.config['SOS_DEBUG'] or 'ALL' in env.config[
                    'SOS_DEBUG']:
                env.log_to_file(
                    'TARGET',
                    f'write signature {self.sig_id} with output {self.output_files}'
                )
        ret = super(RuntimeInfo, self).write()
        if ret is False:
            env.logger.debug(f'Failed to write signature {self.sig_id}')
            return ret
        send_message_to_controller(['step_sig', self.sig_id, ret])
        send_message_to_controller([
            'workflow_sig', 'tracked_files', self.sig_id,
            repr({
                'input_files': [
                    str(f.resolve())
                    for f in self.input_files
                    if isinstance(f, file_target)
                ],
                'dependent_files': [
                    str(f.resolve())
                    for f in self.dependent_files
                    if isinstance(f, file_target)
                ],
                'output_files': [
                    str(f.resolve())
                    for f in self.output_files
                    if isinstance(f, file_target)
                ]
            })
        ])
        return True

    def validate(self):
        '''Check if ofiles and ifiles match signatures recorded in md5file'''
        if not self.sig_id:
            return f'no signature for steps with nested workflow'
        if 'TARGET' in env.config['SOS_DEBUG'] or 'ALL' in env.config[
                'SOS_DEBUG']:
            env.log_to_file('TARGET', f'Validating {self.sig_id}')
        #
        # file not exist?
        sig_files = self.input_files._targets + self.output_files._targets + \
            self.dependent_files._targets
        for x in sig_files:
            if not x.target_exists('any'):
                return f'Missing target {x}'
        #
        sig = request_answer_from_controller(['step_sig', 'get', self.sig_id])
        if not sig:
            return f"No signature found for {self.sig_id}"
        return super(RuntimeInfo, self).validate(sig)



#!/usr/bin/env python3
#
# Copyright (c) Bo Peng and the University of Texas MD Anderson Cancer Center
# Distributed under the terms of the 3-clause BSD License.

import os
import shutil
import subprocess
import unittest
import getpass

from sos._version import __version__
from sos.parser import SoS_Script
from sos.targets import file_target
from sos.utils import env, load_config_files
# if the test is imported under sos/test, test interacive executor
from sos.workflow_executor import Base_Executor


class TestConfig(unittest.TestCase):

    def setUp(self):
        env.reset()
        subprocess.call('sos remove -s', shell=True)
        # self.resetDir('~/.sos')
        self.temp_files = []

    def tearDown(self):
        for f in self.temp_files:
            if file_target(f).exists():
                file_target(f).unlink()

    def touch(self, files):
        '''create temporary files'''
        if isinstance(files, str):
            files = [files]
        #
        for f in files:
            with open(f, 'w') as tmp:
                tmp.write('test')
        #
        self.temp_files.extend(files)

    def resetDir(self, dirname):
        if os.path.isdir(os.path.expanduser(dirname)):
            shutil.rmtree(os.path.expanduser(dirname))
        os.mkdir(os.path.expanduser(dirname))

    def testCommandLine(self):
        '''Test command line arguments'''
        self.assertEqual(
            subprocess.call(
                'sos config -h',
                stderr=subprocess.DEVNULL,
                stdout=subprocess.DEVNULL,
                shell=True), 0)
        self.assertEqual(
            subprocess.call(
                'sos config -g --get',
                stderr=subprocess.DEVNULL,
                stdout=subprocess.DEVNULL,
                shell=True), 0)
        self.assertEqual(
            subprocess.call(
                'sos config',
                stderr=subprocess.DEVNULL,
                stdout=subprocess.DEVNULL,
                shell=True), 0)
        self.assertEqual(
            subprocess.call(
                'sos config --get',
                stderr=subprocess.DEVNULL,
                stdout=subprocess.DEVNULL,
                shell=True), 0)
        self.assertEqual(
            subprocess.call(
                'sos config -g --set a 5',
                stderr=subprocess.DEVNULL,
                stdout=subprocess.DEVNULL,
                shell=True), 0)
        self.assertEqual(
            subprocess.call(
                'sos config --get a',
                stderr=subprocess.DEVNULL,
                stdout=subprocess.DEVNULL,
                shell=True), 0)
        self.assertEqual(
            subprocess.call(
                'sos config -g --unset a',
                stderr=subprocess.DEVNULL,
                stdout=subprocess.DEVNULL,
                shell=True), 0)

    def testGlobalVars(self):
        '''Test SoS defined variables'''
        script = SoS_Script(r"""
[0]
""")
        wf = script.workflow()
        Base_Executor(wf).run(mode='dryrun')
        self.assertEqual(env.sos_dict['SOS_VERSION'], __version__)
        self.assertTrue(isinstance(env.sos_dict['CONFIG'], dict))
        #
        with open('local_cfg.yml', 'w') as cfg:
            cfg.write('my_config: 5')
        #
        Base_Executor(wf, config={'config_file': 'local_cfg.yml'}).run()
        self.assertEqual(env.sos_dict['CONFIG']['my_config'], 5)

    def testConfigSet(self):
        '''Test interpolation of config'''
        self.assertEqual(
            subprocess.call(
                'sos config --set cut 0.5 -c myconfig.yml',
                stderr=subprocess.DEVNULL,
                stdout=subprocess.DEVNULL,
                shell=True), 0)
        load_config_files('myconfig.yml')
        self.assertEqual(env.sos_dict['CONFIG']['cut'], 0.5)
        #
        self.assertEqual(
            subprocess.call(
                'sos config --set cut1 0.5 2 3 -c myconfig.yml',
                stderr=subprocess.DEVNULL,
                stdout=subprocess.DEVNULL,
                shell=True), 0)
        load_config_files('myconfig.yml')
        self.assertEqual(env.sos_dict['CONFIG']['cut1'], [0.5, 2, 3])
        #
        self.assertEqual(
            subprocess.call(
                'sos config --set cut2 a3 -c myconfig.yml',
                stderr=subprocess.DEVNULL,
                stdout=subprocess.DEVNULL,
                shell=True), 0)
        load_config_files('myconfig.yml')
        self.assertEqual(env.sos_dict['CONFIG']['cut2'], 'a3')
        #
        self.assertEqual(
            subprocess.call(
                'sos config --set cut3 a b c -c myconfig.yml',
                stderr=subprocess.DEVNULL,
                stdout=subprocess.DEVNULL,
                shell=True), 0)
        load_config_files('myconfig.yml')
        self.assertEqual(env.sos_dict['CONFIG']['cut3'], ['a', 'b', 'c'])
        #
        self.assertEqual(
            subprocess.call(
                '''sos config --set cut4 "{'A': 123}" -c myconfig.yml''',
                stderr=subprocess.DEVNULL,
                stdout=subprocess.DEVNULL,
                shell=True), 0)
        load_config_files('myconfig.yml')
        self.assertEqual(env.sos_dict['CONFIG']['cut4'], {'A': 123})

    def testInterpolate(self):
        '''Test interpolation of config'''
        self.assertEqual(
            subprocess.call(
                '''sos config --set me '{user_name}@my' -c myconfig.yml''',
                stderr=subprocess.DEVNULL,
                stdout=subprocess.DEVNULL,
                shell=True), 0)
        load_config_files('myconfig.yml')
        self.assertEqual(env.sos_dict['CONFIG']['me'],
                         f'{getpass.getuser().lower()}@my')


if __name__ == '__main__':
    unittest.main()
